#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for operator_utils."""

import os

import itertools

import unittest

import numpy

import sympy

from scipy.sparse import csc_matrix

from openfermion.config import DATA_DIRECTORY
from openfermion.hamiltonians import fermi_hubbard
from openfermion.ops.operators import (FermionOperator, MajoranaOperator,
                                       BosonOperator, QubitOperator,
                                       QuadOperator)
from openfermion.ops.representations import InteractionOperator
from openfermion.transforms.opconversions import (jordan_wigner, bravyi_kitaev)
from openfermion.transforms.repconversions import get_interaction_operator
from openfermion.testing.testing_utils import random_interaction_operator
from openfermion.utils.operator_utils import (count_qubits,
                                              hermitian_conjugated, is_identity,
                                              save_operator, OperatorUtilsError,
                                              is_hermitian, load_operator,
                                              get_file_path)


class OperatorUtilsTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits = 5
        self.majorana_operator = MajoranaOperator((1, 4, 9))
        self.fermion_term = FermionOperator('1^ 2^ 3 4', -3.17)
        self.fermion_operator = self.fermion_term + hermitian_conjugated(
            self.fermion_term)
        self.qubit_operator = jordan_wigner(self.fermion_operator)
        self.interaction_operator = get_interaction_operator(
            self.fermion_operator)

    def test_n_qubits_majorana_operator(self):
        self.assertEqual(self.n_qubits, count_qubits(self.majorana_operator))

    def test_n_qubits_single_fermion_term(self):
        self.assertEqual(self.n_qubits, count_qubits(self.fermion_term))

    def test_n_qubits_fermion_operator(self):
        self.assertEqual(self.n_qubits, count_qubits(self.fermion_operator))

    def test_n_qubits_qubit_operator(self):
        self.assertEqual(self.n_qubits, count_qubits(self.qubit_operator))

    def test_n_qubits_interaction_operator(self):
        self.assertEqual(self.n_qubits, count_qubits(self.interaction_operator))

    def test_n_qubits_bad_type(self):
        with self.assertRaises(TypeError):
            count_qubits('twelve')

    def test_is_identity_unit_fermionoperator(self):
        self.assertTrue(is_identity(FermionOperator(())))

    def test_is_identity_double_of_unit_fermionoperator(self):
        self.assertTrue(is_identity(2. * FermionOperator(())))

    def test_is_identity_unit_bosonoperator(self):
        self.assertTrue(is_identity(BosonOperator(())))

    def test_is_identity_double_of_unit_bosonoperator(self):
        self.assertTrue(is_identity(2. * BosonOperator(())))

    def test_is_identity_unit_quadoperator(self):
        self.assertTrue(is_identity(QuadOperator(())))

    def test_is_identity_double_of_unit_quadoperator(self):
        self.assertTrue(is_identity(2. * QuadOperator(())))

    def test_is_identity_unit_qubitoperator(self):
        self.assertTrue(is_identity(QubitOperator(())))

    def test_is_identity_double_of_unit_qubitoperator(self):
        self.assertTrue(is_identity(QubitOperator((), 2.)))

    def test_not_is_identity_single_term_fermionoperator(self):
        self.assertFalse(is_identity(FermionOperator('1^')))

    def test_not_is_identity_single_term_bosonoperator(self):
        self.assertFalse(is_identity(BosonOperator('1^')))

    def test_not_is_identity_single_term_quadoperator(self):
        self.assertFalse(is_identity(QuadOperator('q1')))

    def test_not_is_identity_single_term_qubitoperator(self):
        self.assertFalse(is_identity(QubitOperator('X1')))

    def test_not_is_identity_zero_bosonoperator(self):
        self.assertFalse(is_identity(BosonOperator()))

    def test_not_is_identity_zero_quadoperator(self):
        self.assertFalse(is_identity(QuadOperator()))

    def test_not_is_identity_zero_qubitoperator(self):
        self.assertFalse(is_identity(QubitOperator()))

    def test_is_identity_bad_type(self):
        with self.assertRaises(TypeError):
            _ = is_identity('eleven')


class HermitianConjugatedTest(unittest.TestCase):

    def test_hermitian_conjugated_qubit_op(self):
        """Test conjugating QubitOperators."""
        op = QubitOperator()
        op_hc = hermitian_conjugated(op)
        correct_op = op
        self.assertEqual(op_hc, correct_op)

        op = QubitOperator('X0 Y1', 2.)
        op_hc = hermitian_conjugated(op)
        correct_op = op
        self.assertEqual(op_hc, correct_op)

        op = QubitOperator('X0 Y1', 2.j)
        op_hc = hermitian_conjugated(op)
        correct_op = QubitOperator('X0 Y1', -2.j)
        self.assertEqual(op_hc, correct_op)

        op = QubitOperator('X0 Y1', 2.) + QubitOperator('Z4 X5 Y7', 3.j)
        op_hc = hermitian_conjugated(op)
        correct_op = (QubitOperator('X0 Y1', 2.) +
                      QubitOperator('Z4 X5 Y7', -3.j))
        self.assertEqual(op_hc, correct_op)

    def test_hermitian_conjugated_qubit_op_consistency(self):
        """Some consistency checks for conjugating QubitOperators."""
        ferm_op = (FermionOperator('1^ 2') + FermionOperator('2 3 4') +
                   FermionOperator('2^ 7 9 11^'))

        # Check that hermitian conjugation commutes with transforms
        self.assertEqual(jordan_wigner(hermitian_conjugated(ferm_op)),
                         hermitian_conjugated(jordan_wigner(ferm_op)))
        self.assertEqual(bravyi_kitaev(hermitian_conjugated(ferm_op)),
                         hermitian_conjugated(bravyi_kitaev(ferm_op)))

    def test_hermitian_conjugated_quad_op(self):
        """Test conjugating QuadOperator."""
        op = QuadOperator()
        op_hc = hermitian_conjugated(op)
        correct_op = op
        self.assertTrue(op_hc == correct_op)

        op = QuadOperator('q0 p1', 2.)
        op_hc = hermitian_conjugated(op)
        correct_op = op
        self.assertTrue(op_hc == correct_op)

        op = QuadOperator('q0 p1', 2.j)
        op_hc = hermitian_conjugated(op)
        correct_op = QuadOperator('q0 p1', -2.j)
        self.assertTrue(op_hc == correct_op)

        op = QuadOperator('q0 p1', 2.) + QuadOperator('q4 q5 p7', 3.j)
        op_hc = hermitian_conjugated(op)
        correct_op = (QuadOperator('q0 p1', 2.) +
                      QuadOperator('q4 q5 p7', -3.j))
        self.assertTrue(op_hc == correct_op)

        op = QuadOperator('q0 p0 q1', 2.) + QuadOperator('q1 p1 p2', 3.j)
        op_hc = hermitian_conjugated(op)
        correct_op = (QuadOperator('p0 q0 q1', 2.) +
                      QuadOperator('p1 q1 p2', -3.j))
        self.assertTrue(op_hc == correct_op)

    def test_hermitian_conjugate_empty(self):
        op = FermionOperator()
        op = hermitian_conjugated(op)
        self.assertEqual(op, FermionOperator())

        op = BosonOperator()
        op = hermitian_conjugated(op)
        self.assertEqual(op, BosonOperator())

    def test_hermitian_conjugate_simple(self):
        op = FermionOperator('1^')
        op_hc = FermionOperator('1')
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

        op = BosonOperator('1^')
        op_hc = BosonOperator('1')
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

    def test_hermitian_conjugate_complex_const(self):
        op = FermionOperator('1^ 3', 3j)
        op_hc = -3j * FermionOperator('3^ 1')
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

        op = BosonOperator('1^ 3', 3j)
        op_hc = -3j * BosonOperator('3^ 1')
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

    def test_hermitian_conjugate_notordered(self):
        op = FermionOperator('1 3^ 3 3^', 3j)
        op_hc = -3j * FermionOperator('3 3^ 3 1^')
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

        op = BosonOperator('1 3^ 3 3^', 3j)
        op_hc = -3j * BosonOperator('3 3^ 3 1^')
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

    def test_hermitian_conjugate_semihermitian(self):
        op = (FermionOperator() + 2j * FermionOperator('1^ 3') +
              FermionOperator('3^ 1') * -2j + FermionOperator('2^ 2', 0.1j))
        op_hc = (FermionOperator() + FermionOperator('1^ 3', 2j) +
                 FermionOperator('3^ 1', -2j) + FermionOperator('2^ 2', -0.1j))
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

        op = (BosonOperator() + 2j * BosonOperator('1^ 3') +
              BosonOperator('3^ 1') * -2j + BosonOperator('2^ 2', 0.1j))
        op_hc = (BosonOperator() + BosonOperator('1^ 3', 2j) +
                 BosonOperator('3^ 1', -2j) + BosonOperator('2^ 2', -0.1j))
        op = hermitian_conjugated(op)
        self.assertEqual(op, op_hc)

    def test_hermitian_conjugated_empty(self):
        op = FermionOperator()
        self.assertEqual(op, hermitian_conjugated(op))
        op = BosonOperator()
        self.assertEqual(op, hermitian_conjugated(op))

    def test_hermitian_conjugated_simple(self):
        op = FermionOperator('0')
        op_hc = FermionOperator('0^')
        self.assertEqual(op_hc, hermitian_conjugated(op))

        op = BosonOperator('0')
        op_hc = BosonOperator('0^')
        self.assertEqual(op_hc, hermitian_conjugated(op))

    def test_hermitian_conjugated_complex_const(self):
        op = FermionOperator('2^ 2', 3j)
        op_hc = FermionOperator('2^ 2', -3j)
        self.assertEqual(op_hc, hermitian_conjugated(op))

        op = BosonOperator('2^ 2', 3j)
        op_hc = BosonOperator('2^ 2', -3j)
        self.assertEqual(op_hc, hermitian_conjugated(op))

    def test_hermitian_conjugated_multiterm(self):
        op = FermionOperator('1^ 2') + FermionOperator('2 3 4')
        op_hc = FermionOperator('2^ 1') + FermionOperator('4^ 3^ 2^')
        self.assertEqual(op_hc, hermitian_conjugated(op))

        op = BosonOperator('1^ 2') + BosonOperator('2 3 4')
        op_hc = BosonOperator('2^ 1') + BosonOperator('4^ 3^ 2^')
        self.assertEqual(op_hc, hermitian_conjugated(op))

    def test_hermitian_conjugated_semihermitian(self):
        op = (FermionOperator() + 2j * FermionOperator('1^ 3') +
              FermionOperator('3^ 1') * -2j + FermionOperator('2^ 2', 0.1j))
        op_hc = (FermionOperator() + FermionOperator('1^ 3', 2j) +
                 FermionOperator('3^ 1', -2j) + FermionOperator('2^ 2', -0.1j))
        self.assertEqual(op_hc, hermitian_conjugated(op))

        op = (BosonOperator() + 2j * BosonOperator('1^ 3') +
              BosonOperator('3^ 1') * -2j + BosonOperator('2^ 2', 0.1j))
        op_hc = (BosonOperator() + BosonOperator('1^ 3', 2j) +
                 BosonOperator('3^ 1', -2j) + BosonOperator('2^ 2', -0.1j))
        self.assertEqual(op_hc, hermitian_conjugated(op))

    def test_hermitian_conjugated_interaction_operator(self):
        for n_orbitals, _ in itertools.product((1, 2, 5), range(5)):
            operator = random_interaction_operator(n_orbitals)
            qubit_operator = jordan_wigner(operator)
            conjugate_operator = hermitian_conjugated(operator)
            conjugate_qubit_operator = jordan_wigner(conjugate_operator)
            assert hermitian_conjugated(qubit_operator) == \
                conjugate_qubit_operator

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            _ = is_hermitian('a')

        with self.assertRaises(TypeError):
            _ = hermitian_conjugated(1)


class IsHermitianTest(unittest.TestCase):

    def test_fermion_operator_zero(self):
        op = FermionOperator()
        self.assertTrue(is_hermitian(op))

    def test_fermion_operator_identity(self):
        op = FermionOperator(())
        self.assertTrue(is_hermitian(op))

    def test_fermion_operator_nonhermitian(self):
        op = FermionOperator('0^ 1 2^ 3')
        self.assertFalse(is_hermitian(op))

    def test_fermion_operator_hermitian(self):
        op = FermionOperator('0^ 1 2^ 3')
        op += FermionOperator('3^ 2 1^ 0')
        self.assertTrue(is_hermitian(op))

        op = fermi_hubbard(2, 2, 1., 1.)
        self.assertTrue(is_hermitian(op))

    def test_boson_operator_zero(self):
        op = BosonOperator()
        self.assertTrue(is_hermitian(op))

    def test_boson_operator_identity(self):
        op = BosonOperator(())
        self.assertTrue(is_hermitian(op))

    def test_boson_operator_nonhermitian(self):
        op = BosonOperator('0^ 1 2^ 3')
        self.assertFalse(is_hermitian(op))

    def test_boson_operator_hermitian(self):
        op = BosonOperator('0^ 1 2^ 3')
        op += BosonOperator('3^ 2 1^ 0')
        self.assertTrue(is_hermitian(op))

    def test_quad_operator_zero(self):
        op = QuadOperator()
        self.assertTrue(is_hermitian(op))

    def test_quad_operator_identity(self):
        op = QuadOperator(())
        self.assertTrue(is_hermitian(op))

    def test_quad_operator_nonhermitian(self):
        op = QuadOperator('q0 p1 q1')
        self.assertFalse(is_hermitian(op))

    def test_quad_operator_hermitian(self):
        op = QuadOperator('q0 p1 q2 p3')
        self.assertTrue(is_hermitian(op))

        op = QuadOperator('q0 p0 q1 p1')
        op += QuadOperator('p0 q0 p1 q1')
        self.assertTrue(is_hermitian(op))

        # TODO: insert bose_hubbard here
        # op = fermi_hubbard(2, 2, 1., 1.)
        # self.assertTrue(is_hermitian(op))

    def test_qubit_operator_zero(self):
        op = QubitOperator()
        self.assertTrue(is_hermitian(op))

    def test_qubit_operator_identity(self):
        op = QubitOperator(())
        self.assertTrue(is_hermitian(op))

    def test_qubit_operator_nonhermitian(self):
        op = QubitOperator('X0 Y2 Z5', 1. + 2.j)
        self.assertFalse(is_hermitian(op))

    def test_qubit_operator_hermitian(self):
        op = QubitOperator('X0 Y2 Z5', 1. + 2.j)
        op += QubitOperator('X0 Y2 Z5', 1. - 2.j)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_zero(self):
        op = numpy.zeros((4, 4))
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_identity(self):
        op = numpy.eye(4)
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_nonhermitian(self):
        op = numpy.arange(16).reshape((4, 4))
        self.assertFalse(is_hermitian(op))
        op = csc_matrix(op)
        self.assertFalse(is_hermitian(op))

    def test_sparse_matrix_and_numpy_array_hermitian(self):
        op = numpy.arange(16, dtype=complex).reshape((4, 4))
        op += 1.j * op
        op += op.T.conj()
        self.assertTrue(is_hermitian(op))
        op = csc_matrix(op)
        self.assertTrue(is_hermitian(op))

    def test_exceptions(self):
        with self.assertRaises(TypeError):
            _ = is_hermitian('a')


class SaveLoadOperatorTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits = 5
        self.fermion_term = FermionOperator('1^ 2^ 3 4', -3.17)
        self.fermion_operator = self.fermion_term + hermitian_conjugated(
            self.fermion_term)
        self.boson_term = BosonOperator('1^ 2^ 3 4', -3.17)
        self.boson_operator = self.boson_term + hermitian_conjugated(
            self.boson_term)
        self.quad_term = QuadOperator('q0 p0 q1 p0 p0', -3.17)
        self.quad_operator = self.quad_term + hermitian_conjugated(
            self.quad_term)
        self.qubit_operator = jordan_wigner(self.fermion_operator)
        self.file_name = "test_file"

        self.bad_operator_filename = 'bad_file.data'
        bad_op = "A:\nB"
        with open(os.path.join(DATA_DIRECTORY, self.bad_operator_filename),
                  'w') as fid:
            fid.write(bad_op)

    def tearDown(self):
        file_path = os.path.join(DATA_DIRECTORY, self.file_name + '.data')
        if os.path.isfile(file_path):
            os.remove(file_path)
        file_path = os.path.join(DATA_DIRECTORY, self.bad_operator_filename)
        if os.path.isfile(file_path):
            os.remove(file_path)

    def test_save_sympy_plaintext(self):
        operator = FermionOperator('1^', sympy.Symbol('x'))
        with self.assertRaises(TypeError):
            save_operator(operator, self.file_name, plain_text=True)

    def test_raises_error_sympy(self):
        operator = FermionOperator('1^', sympy.Symbol('x'))
        with self.assertRaises(TypeError):
            save_operator(operator, self.file_name, plain_text=False)

    def test_save_and_load_fermion_operators(self):
        save_operator(self.fermion_operator, self.file_name)
        loaded_fermion_operator = load_operator(self.file_name)
        self.assertEqual(self.fermion_operator,
                         loaded_fermion_operator,
                         msg=str(self.fermion_operator -
                                 loaded_fermion_operator))

    def test_save_and_load_fermion_operators_readably(self):
        save_operator(self.fermion_operator, self.file_name, plain_text=True)
        loaded_fermion_operator = load_operator(self.file_name, plain_text=True)
        self.assertTrue(self.fermion_operator == loaded_fermion_operator)

    def test_save_and_load_boson_operators(self):
        save_operator(self.boson_operator, self.file_name)
        loaded_boson_operator = load_operator(self.file_name)
        self.assertEqual(self.boson_operator.terms,
                         loaded_boson_operator.terms,
                         msg=str(self.boson_operator - loaded_boson_operator))

    def test_save_and_load_boson_operators_readably(self):
        save_operator(self.boson_operator, self.file_name, plain_text=True)
        loaded_boson_operator = load_operator(self.file_name, plain_text=True)
        self.assertTrue(self.boson_operator == loaded_boson_operator)

    def test_save_and_load_quad_operators(self):
        save_operator(self.quad_operator, self.file_name)
        loaded_quad_operator = load_operator(self.file_name)
        self.assertEqual(self.quad_operator.terms, loaded_quad_operator.terms)

    def test_save_and_load_quad_operators_readably(self):
        save_operator(self.quad_operator, self.file_name, plain_text=True)
        loaded_quad_operator = load_operator(self.file_name, plain_text=True)
        self.assertTrue(self.quad_operator == loaded_quad_operator)

    def test_save_and_load_qubit_operators(self):
        save_operator(self.qubit_operator, self.file_name)
        loaded_qubit_operator = load_operator(self.file_name)
        self.assertTrue(self.qubit_operator == loaded_qubit_operator)

    def test_save_and_load_qubit_operators_readably(self):
        save_operator(self.qubit_operator, self.file_name, plain_text=True)
        loaded_qubit_operator = load_operator(self.file_name, plain_text=True)
        self.assertEqual(self.qubit_operator, loaded_qubit_operator)

    def test_load_bad_operator(self):
        with self.assertRaises(TypeError):
            load_operator(self.bad_operator_filename, plain_text=True)

    def test_save_readably(self):
        save_operator(self.fermion_operator, self.file_name, plain_text=True)
        file_path = os.path.join(DATA_DIRECTORY, self.file_name + '.data')
        with open(file_path, "r") as f:
            self.assertEqual(
                f.read(), "\n".join([
                    "FermionOperator:", "-3.17 [1^ 2^ 3 4] +",
                    "-3.17 [4^ 3^ 2 1]"
                ]))

    def test_save_no_filename_operator_utils_error(self):
        with self.assertRaises(OperatorUtilsError):
            save_operator(self.fermion_operator)

    def test_basic_save(self):
        save_operator(self.fermion_operator, self.file_name)

    def test_save_interaction_operator_not_implemented(self):
        constant = 100.0
        one_body = numpy.zeros((self.n_qubits, self.n_qubits), float)
        two_body = numpy.zeros(
            (self.n_qubits, self.n_qubits, self.n_qubits, self.n_qubits), float)
        one_body[1, 1] = 10.0
        two_body[1, 2, 3, 4] = 12.0
        interaction_operator = InteractionOperator(constant, one_body, two_body)
        with self.assertRaises(NotImplementedError):
            save_operator(interaction_operator, self.file_name)

    def test_save_on_top_of_existing_operator_utils_error(self):
        save_operator(self.fermion_operator, self.file_name)
        with self.assertRaises(OperatorUtilsError):
            save_operator(self.fermion_operator, self.file_name)

    def test_save_on_top_of_existing_operator_error_with_explicit_flag(self):
        save_operator(self.fermion_operator, self.file_name)
        with self.assertRaises(OperatorUtilsError):
            save_operator(self.fermion_operator,
                          self.file_name,
                          allow_overwrite=False)

    def test_overwrite_flag_save_on_top_of_existing_operator(self):
        save_operator(self.fermion_operator, self.file_name)
        save_operator(self.fermion_operator,
                      self.file_name,
                      allow_overwrite=True)
        fermion_operator = load_operator(self.file_name)

        self.assertEqual(fermion_operator, self.fermion_operator)

    def test_load_bad_type(self):
        with self.assertRaises(TypeError):
            _ = load_operator('bad_type_operator')

    def test_save_bad_type(self):
        with self.assertRaises(TypeError):
            save_operator('ping', 'somewhere')


class GetFileDirTest(unittest.TestCase):

    def setUp(self):
        self.filename = 'foo'
        self.datadirname = 'data'

    def test_file_load(self):
        """Test if file name is acquired correctly"""
        filepath = get_file_path(self.filename, None)
        self.assertEqual(filepath,
                         DATA_DIRECTORY + '/' + self.filename + '.data')

        filepath = get_file_path(self.filename, self.datadirname)
        self.assertEqual(filepath,
                         self.datadirname + '/' + self.filename + '.data')


#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import unittest

import numpy

from openfermion.testing.testing_utils import (random_quadratic_hamiltonian,
                                               random_unitary_matrix)

from openfermion.linalg.givens_rotations import (
    double_givens_rotate, fermionic_gaussian_decomposition, swap_columns,
    givens_matrix_elements, givens_rotate, givens_decomposition, swap_rows,
    givens_decomposition_square)


class SwapTests(unittest.TestCase):

    def test_swap_rows(self):
        A = numpy.array([0, 1])
        swap_rows(A, 0, 1)
        self.assertEqual(A[0], 1)
        self.assertEqual(A[1], 0)
        A = numpy.array([[0, 1], [2, 3]])
        swap_rows(A, 0, 1)
        self.assertEqual(A[0, 0], 2)
        self.assertEqual(A[0, 1], 3)
        self.assertEqual(A[1, 0], 0)
        self.assertEqual(A[1, 1], 1)

    def test_swap_columns(self):
        A = numpy.array([0, 1])
        swap_columns(A, 0, 1)
        self.assertEqual(A[0], 1)
        self.assertEqual(A[1], 0)
        A = numpy.array([[0, 1], [2, 3]])
        swap_columns(A, 0, 1)
        self.assertEqual(A[0, 0], 1)
        self.assertEqual(A[0, 1], 0)
        self.assertEqual(A[1, 0], 3)
        self.assertEqual(A[1, 1], 2)


class GivensMatrixElementsTest(unittest.TestCase):

    def setUp(self):
        self.num_test_repetitions = 5

    def test_already_zero(self):
        """Test when some entries are already zero."""
        # Test when left entry is zero
        v = numpy.array([0., numpy.random.randn()])
        G = givens_matrix_elements(v[0], v[1])
        self.assertAlmostEqual(G.dot(v)[0], 0.)
        # Test when right entry is zero
        v = numpy.array([numpy.random.randn(), 0.])
        G = givens_matrix_elements(v[0], v[1])
        self.assertAlmostEqual(G.dot(v)[0], 0.)

    def test_real(self):
        """Test that the procedure works for real numbers."""
        for _ in range(self.num_test_repetitions):
            v = numpy.random.randn(2)
            G_left = givens_matrix_elements(v[0], v[1], which='left')
            G_right = givens_matrix_elements(v[0], v[1], which='right')
            self.assertAlmostEqual(G_left.dot(v)[0], 0.)
            self.assertAlmostEqual(G_right.dot(v)[1], 0.)

    def test_approximately_real(self):
        """Test that the procedure throws out small imaginary components."""
        for _ in range(self.num_test_repetitions):
            v = numpy.random.randn(2) + 1.j * 1e-14
            G_left = givens_matrix_elements(v[0], v[1], which='left')
            G_right = givens_matrix_elements(v[0], v[1], which='right')
            self.assertAlmostEqual(G_left[0, 0], G_left[1, 1])
            self.assertAlmostEqual(G_right[0, 0], G_right[1, 1])
            self.assertAlmostEqual(G_left.dot(v)[0], 0.)
            self.assertAlmostEqual(G_right.dot(v)[1], 0.)

    def test_bad_input(self):
        """Test bad input."""
        with self.assertRaises(ValueError):
            v = numpy.random.randn(2)
            _ = givens_matrix_elements(v[0], v[1], which='a')


class GivensRotateTest(unittest.TestCase):

    def test_bad_input(self):
        """Test bad input."""
        with self.assertRaises(ValueError):
            v = numpy.random.randn(2)
            G = givens_matrix_elements(v[0], v[1])
            givens_rotate(v, G, 0, 1, which='a')


class DoubleGivensRotateTest(unittest.TestCase):

    def test_odd_dimension(self):
        """Test that it raises an error for odd-dimensional input."""
        A = numpy.random.randn(3, 3)
        v = numpy.random.randn(2)
        G = givens_matrix_elements(v[0], v[1])
        with self.assertRaises(ValueError):
            double_givens_rotate(A, G, 0, 1, which='row')
        with self.assertRaises(ValueError):
            double_givens_rotate(A, G, 0, 1, which='col')

    def test_bad_input(self):
        """Test bad input."""
        A = numpy.random.randn(3, 3)
        v = numpy.random.randn(2)
        G = givens_matrix_elements(v[0], v[1])
        with self.assertRaises(ValueError):
            double_givens_rotate(A, G, 0, 1, which='a')


class GivensDecompositionTest(unittest.TestCase):

    def setUp(self):
        self.test_dimensions = [(3, 4), (3, 5), (3, 6), (3, 7), (3, 8), (3, 9),
                                (4, 7), (4, 8), (4, 9)]

    def test_forced_insertion(self):
        Q = numpy.zeros([2, 4])
        Q[0, 0] = Q[1, 1] = 1
        givens_rotations, _, _ = givens_decomposition(Q, always_insert=True)
        assert len(givens_rotations) == 3

    def test_forced_insertion_square(self):
        Q = numpy.eye(4)
        givens_rotations, _ = givens_decomposition_square(Q, always_insert=True)
        assert len(givens_rotations) == 5

    def test_main_procedure(self):
        for m, n in self.test_dimensions:
            # Obtain a random matrix of orthonormal rows
            Q = random_unitary_matrix(n)
            Q = Q[:m, :]

            # Get Givens decomposition of Q
            givens_rotations, V, diagonal = givens_decomposition(Q)

            # Compute U
            U = numpy.eye(n, dtype=complex)
            for parallel_set in givens_rotations:
                combined_givens = numpy.eye(n, dtype=complex)
                for i, j, theta, phi in reversed(parallel_set):
                    c = numpy.cos(theta)
                    s = numpy.sin(theta)
                    phase = numpy.exp(1.j * phi)
                    G = numpy.array([[c, -phase * s], [s, phase * c]],
                                    dtype=complex)
                    givens_rotate(combined_givens, G, i, j)
                U = combined_givens.dot(U)

            # Compute V * Q * U^\dagger
            W = V.dot(Q.dot(U.T.conj()))

            # Construct the diagonal matrix
            D = numpy.zeros((m, n), dtype=complex)
            D[numpy.diag_indices(m)] = diagonal

            # Assert that W and D are the same
            for i in range(m):
                for j in range(n):
                    self.assertAlmostEqual(D[i, j], W[i, j])

    def test_real_numbers(self):
        for m, n in self.test_dimensions:
            # Obtain a random real matrix of orthonormal rows
            Q = random_unitary_matrix(n, real=True)
            Q = Q[:m, :]

            # Get Givens decomposition of Q
            givens_rotations, V, diagonal = givens_decomposition(Q)

            # Compute U
            U = numpy.eye(n, dtype=complex)
            for parallel_set in givens_rotations:
                combined_givens = numpy.eye(n, dtype=complex)
                for i, j, theta, phi in reversed(parallel_set):
                    c = numpy.cos(theta)
                    s = numpy.sin(theta)
                    phase = numpy.exp(1.j * phi)
                    G = numpy.array([[c, -phase * s], [s, phase * c]],
                                    dtype=complex)
                    givens_rotate(combined_givens, G, i, j)
                U = combined_givens.dot(U)

            # Compute V * Q * U^\dagger
            W = V.dot(Q.dot(U.T.conj()))

            # Construct the diagonal matrix
            D = numpy.zeros((m, n), dtype=complex)
            D[numpy.diag_indices(m)] = diagonal

            # Assert that W and D are the same
            for i in range(m):
                for j in range(n):
                    self.assertAlmostEqual(D[i, j], W[i, j])

    def test_bad_dimensions(self):
        m, n = (3, 2)

        # Obtain a random matrix of orthonormal rows
        Q = random_unitary_matrix(m)
        Q = Q[:m, :]
        Q = Q[:m, :n]

        with self.assertRaises(ValueError):
            _ = givens_decomposition(Q)

    def test_identity(self):
        n = 3
        Q = numpy.eye(n, dtype=complex)
        givens_rotations, V, diagonal = givens_decomposition(Q)

        # V should be the identity
        identity = numpy.eye(n, dtype=complex)
        for i in range(n):
            for j in range(n):
                self.assertAlmostEqual(V[i, j], identity[i, j])

        # There should be no Givens rotations
        self.assertEqual(givens_rotations, list())

        # The diagonal should be ones
        for d in diagonal:
            self.assertAlmostEqual(d, 1.)

    def test_antidiagonal(self):
        m, n = (3, 3)
        Q = numpy.zeros((m, n), dtype=complex)
        Q[0, 2] = 1.
        Q[1, 1] = 1.
        Q[2, 0] = 1.
        givens_rotations, V, diagonal = givens_decomposition(Q)

        # There should be no Givens rotations
        self.assertEqual(givens_rotations, list())

        # VQ should equal the diagonal
        VQ = V.dot(Q)
        D = numpy.zeros((m, n), dtype=complex)
        D[numpy.diag_indices(m)] = diagonal
        for i in range(n):
            for j in range(n):
                self.assertAlmostEqual(VQ[i, j], D[i, j])

    def test_square(self):
        m, n = (3, 3)

        # Obtain a random matrix of orthonormal rows
        Q = random_unitary_matrix(n)
        Q = Q[:m, :]
        Q = Q[:m, :]

        # Get Givens decomposition of Q
        givens_rotations, V, diagonal = givens_decomposition(Q)

        # There should be no Givens rotations
        self.assertEqual(givens_rotations, list())

        # Compute V * Q * U^\dagger
        W = V.dot(Q)

        # Construct the diagonal matrix
        D = numpy.zeros((m, n), dtype=complex)
        D[numpy.diag_indices(m)] = diagonal

        # Assert that W and D are the same
        for i in range(m):
            for j in range(n):
                self.assertAlmostEqual(D[i, j], W[i, j])


class FermionicGaussianDecompositionTest(unittest.TestCase):

    def setUp(self):
        self.test_dimensions = [3, 4, 5, 6, 7, 8, 9]

    def test_main_procedure(self):
        for n in self.test_dimensions:
            # Obtain a random quadratic Hamiltonian
            quadratic_hamiltonian = random_quadratic_hamiltonian(n)

            # Get the diagonalizing transformation
            _, transformation_matrix, _ = (
                quadratic_hamiltonian.diagonalizing_bogoliubov_transform())
            left_block = transformation_matrix[:, :n]
            right_block = transformation_matrix[:, n:]
            lower_unitary = numpy.empty((n, 2 * n), dtype=complex)
            lower_unitary[:, :n] = numpy.conjugate(right_block)
            lower_unitary[:, n:] = numpy.conjugate(left_block)

            # Get fermionic Gaussian decomposition of lower_unitary
            decomposition, left_decomposition, diagonal, left_diagonal = (
                fermionic_gaussian_decomposition(lower_unitary))

            # Compute left_unitary
            left_unitary = numpy.eye(n, dtype=complex)
            for parallel_set in left_decomposition:
                combined_op = numpy.eye(n, dtype=complex)
                for op in reversed(parallel_set):
                    i, j, theta, phi = op
                    c = numpy.cos(theta)
                    s = numpy.sin(theta)
                    phase = numpy.exp(1.j * phi)
                    givens_rotation = numpy.array(
                        [[c, -phase * s], [s, phase * c]], dtype=complex)
                    givens_rotate(combined_op, givens_rotation, i, j)
                left_unitary = combined_op.dot(left_unitary)
            for i in range(n):
                left_unitary[i] *= left_diagonal[i]
            left_unitary = left_unitary.T
            for i in range(n):
                left_unitary[i] *= diagonal[i]

            # Check that left_unitary zeroes out the correct entries of
            # lower_unitary
            product = left_unitary.dot(lower_unitary)
            for i in range(n - 1):
                for j in range(n - 1 - i):
                    self.assertAlmostEqual(product[i, j], 0.)

            # Compute right_unitary
            right_unitary = numpy.eye(2 * n, dtype=complex)
            for parallel_set in decomposition:
                combined_op = numpy.eye(2 * n, dtype=complex)
                for op in reversed(parallel_set):
                    if op == 'pht':
                        swap_rows(combined_op, n - 1, 2 * n - 1)
                    else:
                        i, j, theta, phi = op
                        c = numpy.cos(theta)
                        s = numpy.sin(theta)
                        phase = numpy.exp(1.j * phi)
                        givens_rotation = numpy.array(
                            [[c, -phase * s], [s, phase * c]], dtype=complex)
                        double_givens_rotate(combined_op, givens_rotation, i, j)
                right_unitary = combined_op.dot(right_unitary)

            # Compute left_unitary * lower_unitary * right_unitary^\dagger
            product = left_unitary.dot(lower_unitary.dot(
                right_unitary.T.conj()))

            # Construct the diagonal matrix
            diag = numpy.zeros((n, 2 * n), dtype=complex)
            diag[range(n), range(n, 2 * n)] = diagonal

            # Assert that W and D are the same
            for i in numpy.ndindex((n, 2 * n)):
                self.assertAlmostEqual(diag[i], product[i])

    def test_bad_dimensions(self):
        n, p = (3, 7)
        rand_mat = numpy.random.randn(n, p)
        with self.assertRaises(ValueError):
            _ = fermionic_gaussian_decomposition(rand_mat)

    def test_bad_constraints(self):
        n = 3
        ones_mat = numpy.ones((n, 2 * n))
        with self.assertRaises(ValueError):
            _ = fermionic_gaussian_decomposition(ones_mat)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""This module constructs Hamiltonians for the Fermi- and Bose-Hubbard models.
"""

from openfermion.ops.operators import BosonOperator, FermionOperator
from openfermion.utils.indexing import down_index, up_index
from openfermion.hamiltonians.special_operators import number_operator


def fermi_hubbard(x_dimension,
                  y_dimension,
                  tunneling,
                  coulomb,
                  chemical_potential=0.,
                  magnetic_field=0.,
                  periodic=True,
                  spinless=False,
                  particle_hole_symmetry=False):
    r"""Return symbolic representation of a Fermi-Hubbard Hamiltonian.

    The idea of this model is that some fermions move around on a grid and the
    energy of the model depends on where the fermions are.
    The Hamiltonians of this model live on a grid of dimensions
    `x_dimension` x `y_dimension`.
    The grid can have periodic boundary conditions or not.
    In the standard Fermi-Hubbard model (which we call the "spinful" model),
    there is room for an "up" fermion and a "down" fermion at each site on the
    grid. In this model, there are a total of `2N` spin-orbitals,
    where `N = x_dimension * y_dimension` is the number of sites.
    In the spinless model, there is only one spin-orbital per site
    for a total of `N`.

    The Hamiltonian for the spinful model has the form

    .. math::

        \begin{align}
        H = &- t \sum_{\langle i,j \rangle} \sum_{\sigma}
                     (a^\dagger_{i, \sigma} a_{j, \sigma} +
                      a^\dagger_{j, \sigma} a_{i, \sigma})
             + U \sum_{i} a^\dagger_{i, \uparrow} a_{i, \uparrow}
                         a^\dagger_{i, \downarrow} a_{i, \downarrow}
            \\
            &- \mu \sum_i \sum_{\sigma} a^\dagger_{i, \sigma} a_{i, \sigma}
             - h \sum_i (a^\dagger_{i, \uparrow} a_{i, \uparrow} -
                       a^\dagger_{i, \downarrow} a_{i, \downarrow})
        \end{align}

    where

        - The indices :math:`\langle i, j \rangle` run over pairs
          :math:`i` and :math:`j` of sites that are connected to each other
          in the grid
        - :math:`\sigma \in \{\uparrow, \downarrow\}` is the spin
        - :math:`t` is the tunneling amplitude
        - :math:`U` is the Coulomb potential
        - :math:`\mu` is the chemical potential
        - :math:`h` is the magnetic field

    One can also construct the Hamiltonian for the spinless model, which
    has the form

    .. math::

        H = - t \sum_{\langle i, j \rangle} (a^\dagger_i a_j + a^\dagger_j a_i)
            + U \sum_{\langle i, j \rangle} a^\dagger_i a_i a^\dagger_j a_j
            - \mu \sum_i a_i^\dagger a_i.

    Args:
        x_dimension (int): The width of the grid.
        y_dimension (int): The height of the grid.
        tunneling (float): The tunneling amplitude :math:`t`.
        coulomb (float): The attractive local interaction strength :math:`U`.
        chemical_potential (float, optional): The chemical potential
            :math:`\mu` at each site. Default value is 0.
        magnetic_field (float, optional): The magnetic field :math:`h`
            at each site. Default value is 0. Ignored for the spinless case.
        periodic (bool, optional): If True, add periodic boundary conditions.
            Default is True.
        spinless (bool, optional): If True, return a spinless Fermi-Hubbard
            model. Default is False.
        particle_hole_symmetry (bool, optional): If False, the repulsion
            term corresponds to:

            .. math::

                U \sum_{k=1}^{N-1} a_k^\dagger a_k a_{k+1}^\dagger a_{k+1}

            If True, the repulsion term is replaced by:

            .. math::

                U \sum_{k=1}^{N-1} (a_k^\dagger a_k - \frac12)
                                   (a_{k+1}^\dagger a_{k+1} - \frac12)

            which is unchanged under a particle-hole transformation.
            Default is False

    Returns:
        hubbard_model: An instance of the FermionOperator class.
    """
    if spinless:
        return _spinless_fermi_hubbard_model(x_dimension, y_dimension,
                                             tunneling, coulomb,
                                             chemical_potential, magnetic_field,
                                             periodic, particle_hole_symmetry)
    else:
        return _spinful_fermi_hubbard_model(x_dimension, y_dimension, tunneling,
                                            coulomb, chemical_potential,
                                            magnetic_field, periodic,
                                            particle_hole_symmetry)


def _spinful_fermi_hubbard_model(x_dimension, y_dimension, tunneling, coulomb,
                                 chemical_potential, magnetic_field, periodic,
                                 particle_hole_symmetry):

    # Initialize operator.
    n_sites = x_dimension * y_dimension
    n_spin_orbitals = 2 * n_sites
    hubbard_model = FermionOperator()

    # Loop through sites and add terms.
    for site in range(n_sites):

        # Get indices of right and bottom neighbors
        right_neighbor = _right_neighbor(site, x_dimension, y_dimension,
                                         periodic)
        bottom_neighbor = _bottom_neighbor(site, x_dimension, y_dimension,
                                           periodic)

        # Avoid double-counting edges when one of the dimensions is 2
        # and the system is periodic
        if x_dimension == 2 and periodic and site % 2 == 1:
            right_neighbor = None
        if y_dimension == 2 and periodic and site >= x_dimension:
            bottom_neighbor = None

        # Add hopping terms with neighbors to the right and bottom.
        if right_neighbor is not None:
            hubbard_model += _hopping_term(up_index(site),
                                           up_index(right_neighbor), -tunneling)
            hubbard_model += _hopping_term(down_index(site),
                                           down_index(right_neighbor),
                                           -tunneling)
        if bottom_neighbor is not None:
            hubbard_model += _hopping_term(up_index(site),
                                           up_index(bottom_neighbor),
                                           -tunneling)
            hubbard_model += _hopping_term(down_index(site),
                                           down_index(bottom_neighbor),
                                           -tunneling)

        # Add local pair Coulomb interaction terms.
        hubbard_model += _coulomb_interaction_term(n_spin_orbitals,
                                                   up_index(site),
                                                   down_index(site), coulomb,
                                                   particle_hole_symmetry)

        # Add chemical potential and magnetic field terms.
        hubbard_model += number_operator(n_spin_orbitals, up_index(site),
                                         -chemical_potential - magnetic_field)
        hubbard_model += number_operator(n_spin_orbitals, down_index(site),
                                         -chemical_potential + magnetic_field)

    return hubbard_model


def _spinless_fermi_hubbard_model(x_dimension, y_dimension, tunneling, coulomb,
                                  chemical_potential, magnetic_field, periodic,
                                  particle_hole_symmetry):

    # Initialize operator.
    n_sites = x_dimension * y_dimension
    hubbard_model = FermionOperator()

    # Loop through sites and add terms.
    for site in range(n_sites):

        # Get indices of right and bottom neighbors
        right_neighbor = _right_neighbor(site, x_dimension, y_dimension,
                                         periodic)
        bottom_neighbor = _bottom_neighbor(site, x_dimension, y_dimension,
                                           periodic)

        # Avoid double-counting edges when one of the dimensions is 2
        # and the system is periodic
        if x_dimension == 2 and periodic and site % 2 == 1:
            right_neighbor = None
        if y_dimension == 2 and periodic and site >= x_dimension:
            bottom_neighbor = None

        # Add terms that couple with neighbors to the right and bottom.
        if right_neighbor is not None:
            # Add hopping term
            hubbard_model += _hopping_term(site, right_neighbor, -tunneling)
            # Add local Coulomb interaction term
            hubbard_model += _coulomb_interaction_term(n_sites, site,
                                                       right_neighbor, coulomb,
                                                       particle_hole_symmetry)
        if bottom_neighbor is not None:
            # Add hopping term
            hubbard_model += _hopping_term(site, bottom_neighbor, -tunneling)
            # Add local Coulomb interaction term
            hubbard_model += _coulomb_interaction_term(n_sites, site,
                                                       bottom_neighbor, coulomb,
                                                       particle_hole_symmetry)

        # Add chemical potential. The magnetic field doesn't contribute.
        hubbard_model += number_operator(n_sites, site, -chemical_potential)

    return hubbard_model


def bose_hubbard(x_dimension,
                 y_dimension,
                 tunneling,
                 interaction,
                 chemical_potential=0.,
                 dipole=0.,
                 periodic=True):
    r"""Return symbolic representation of a Bose-Hubbard Hamiltonian.

    In this model, bosons move around on a lattice, and the
    energy of the model depends on where the bosons are.

    The lattice is described by a 2D grid, with dimensions
    `x_dimension` x `y_dimension`. It is also possible to specify
    if the grid has periodic boundary conditions or not.

    The Hamiltonian for the Bose-Hubbard model has the form

    .. math::

        H = - t \sum_{\langle i, j \rangle} (b_i^\dagger b_j + b_j^\dagger b_i)
         + V \sum_{\langle i, j \rangle} b_i^\dagger b_i b_j^\dagger b_j
         + \frac{U}{2} \sum_i b_i^\dagger b_i (b_i^\dagger b_i - 1)
         - \mu \sum_i b_i^\dagger b_i.

    where

        - The indices :math:`\langle i, j \rangle` run over pairs
          :math:`i` and :math:`j` of nodes that are connected to each other
          in the grid
        - :math:`t` is the tunneling amplitude
        - :math:`U` is the on-site interaction potential
        - :math:`\mu` is the chemical potential
        - :math:`V` is the dipole or nearest-neighbour interaction potential

    Args:
        x_dimension (int): The width of the grid.
        y_dimension (int): The height of the grid.
        tunneling (float): The tunneling amplitude :math:`t`.
        interaction (float): The attractive local interaction
            strength :math:`U`.
        chemical_potential (float, optional): The chemical potential
            :math:`\mu` at each site. Default value is 0.
        periodic (bool, optional): If True, add periodic boundary conditions.
            Default is True.
        dipole (float): The attractive dipole interaction strength :math:`V`.

    Returns:
        bose_hubbard_model: An instance of the BosonOperator class.
    """

    # Initialize operator.
    n_sites = x_dimension * y_dimension
    hubbard_model = BosonOperator()

    # Loop through sites and add terms.
    for site in range(n_sites):

        # Get indices of right and bottom neighbors
        right_neighbor = _right_neighbor(site, x_dimension, y_dimension,
                                         periodic)
        bottom_neighbor = _bottom_neighbor(site, x_dimension, y_dimension,
                                           periodic)

        # Avoid double-counting edges when one of the dimensions is 2
        # and the system is periodic
        if x_dimension == 2 and periodic and site % 2 == 1:
            right_neighbor = None
        if y_dimension == 2 and periodic and site >= x_dimension:
            bottom_neighbor = None

        # Add terms that couple with neighbors to the right and bottom.
        if right_neighbor is not None:
            # Add hopping term
            hubbard_model += _hopping_term(site,
                                           right_neighbor,
                                           -tunneling,
                                           bosonic=True)
            # Add local Coulomb interaction term
            hubbard_model += _coulomb_interaction_term(
                n_sites,
                site,
                right_neighbor,
                dipole,
                particle_hole_symmetry=False,
                bosonic=True)
        if bottom_neighbor is not None:
            # Add hopping term
            hubbard_model += _hopping_term(site,
                                           bottom_neighbor,
                                           -tunneling,
                                           bosonic=True)
            # Add local Coulomb interaction term
            hubbard_model += _coulomb_interaction_term(
                n_sites,
                site,
                bottom_neighbor,
                dipole,
                particle_hole_symmetry=False,
                bosonic=True)

        # Add on-site interaction.
        hubbard_model += (
            number_operator(n_sites, site, 0.5 * interaction, parity=1) *
            (number_operator(n_sites, site, parity=1) - BosonOperator(())))

        # Add chemical potential.
        hubbard_model += number_operator(n_sites,
                                         site,
                                         -chemical_potential,
                                         parity=1)

    return hubbard_model


def _hopping_term(i, j, coefficient, bosonic=False):
    op_class = BosonOperator if bosonic else FermionOperator
    hopping_term = op_class(((i, 1), (j, 0)), coefficient)
    hopping_term += op_class(((j, 1), (i, 0)), coefficient.conjugate())
    return hopping_term


def _coulomb_interaction_term(n_sites,
                              i,
                              j,
                              coefficient,
                              particle_hole_symmetry,
                              bosonic=False):
    op_class = BosonOperator if bosonic else FermionOperator
    number_operator_i = number_operator(n_sites, i, parity=2 * bosonic - 1)
    number_operator_j = number_operator(n_sites, j, parity=2 * bosonic - 1)
    if particle_hole_symmetry:
        number_operator_i -= op_class((), 0.5)
        number_operator_j -= op_class((), 0.5)
    return coefficient * number_operator_i * number_operator_j


def _right_neighbor(site, x_dimension, y_dimension, periodic):
    if x_dimension == 1:
        return None
    if (site + 1) % x_dimension == 0:
        if periodic:
            return site + 1 - x_dimension
        else:
            return None
    return site + 1


def _bottom_neighbor(site, x_dimension, y_dimension, periodic):
    if y_dimension == 1:
        return None
    if site + x_dimension + 1 > x_dimension * y_dimension:
        if periodic:
            return site + x_dimension - x_dimension * y_dimension
        else:
            return None
    return site + x_dimension



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Optional, Sequence
import cirq
from openfermion.ops.operators import QubitOperator
from openfermion.utils.operator_utils import count_qubits


def _qubit_operator_term_to_pauli_string(term: dict, qubits: Sequence[cirq.Qid]
                                        ) -> cirq.PauliString:
    """
    Convert term of QubitOperator to a PauliString.

    Args:
        term (dict): QubitOperator term.
        qubits (list): List of qubit names.
    Returns:
        pauli_string (PauliString): cirq PauliString object.
    """
    ind_ops, coeff = term

    return cirq.PauliString(dict((qubits[ind], op) for ind, op in ind_ops),
                            coeff)


def qubit_operator_to_pauli_sum(operator: QubitOperator,
                                qubits: Optional[Sequence[cirq.Qid]] = None
                               ) -> cirq.PauliSum:
    """
    Convert QubitOperator to a sum of PauliString.

    Args:
        operator (QubitOperator): operator to convert.
        qubits (List): Optional list of qubit names.
            If `None` a list of `cirq.LineQubit` of length number of qubits
            in operator is created.

    Returns:
        pauli_sum (PauliSum): cirq PauliSum object.

    Raises:
        TypeError: if qubit_op is not a QubitOpertor.
    """
    if not isinstance(operator, QubitOperator):
        raise TypeError('Input must be a QubitOperator.')

    if qubits is None:
        qubits = cirq.LineQubit.range(count_qubits(operator))

    pauli_sum = cirq.PauliSum()
    for pauli in operator.terms.items():
        pauli_sum += _qubit_operator_term_to_pauli_string(pauli, qubits)

    return pauli_sum



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for doci_hamiltonian.py."""
import os
import unittest

import numpy

from openfermion.config import EQ_TOLERANCE
from openfermion.chem.molecular_data import MolecularData
from openfermion.config import DATA_DIRECTORY
from openfermion.transforms import jordan_wigner
from openfermion.linalg import get_sparse_operator
from openfermion.ops.representations.doci_hamiltonian import (
    DOCIHamiltonian, get_tensors_from_doci, get_projected_integrals_from_doci,
    get_doci_from_integrals)


class IntegralTransformsTest(unittest.TestCase):

    def setUp(self):
        self.geometry = [('H', (0., 0., 0.)), ('H', (0., 0., 0.7414))]
        self.basis = 'sto-3g'
        self.multiplicity = 1
        self.filename = os.path.join(DATA_DIRECTORY, 'H2_sto-3g_singlet_0.7414')
        self.molecule = MolecularData(self.geometry,
                                      self.basis,
                                      self.multiplicity,
                                      filename=self.filename)
        self.molecule.load()

    def test_integrals_self_inverse(self):
        hc, hr1, hr2 = get_doci_from_integrals(self.molecule.one_body_integrals,
                                               self.molecule.two_body_integrals)
        proj_one_body, proj_two_body = get_projected_integrals_from_doci(
            hc, hr1, hr2)
        hc_test, hr1_test, hr2_test = get_doci_from_integrals(
            proj_one_body, proj_two_body)
        self.assertTrue(numpy.allclose(hc, hc_test))
        self.assertTrue(numpy.allclose(hr1, hr1_test))
        print(hr2)
        print(hr2_test)
        self.assertTrue(numpy.allclose(hr2, hr2_test))

    def test_integrals_to_doci(self):
        one_body_integrals = self.molecule.one_body_integrals
        two_body_integrals = self.molecule.two_body_integrals
        hc, hr1, hr2 = get_doci_from_integrals(one_body_integrals,
                                               two_body_integrals)
        self.assertEqual(hc.shape[0], 2)
        self.assertEqual(hr1.shape[0], 2)
        self.assertEqual(hr2.shape[0], 2)

        for p in range(2):
            self.assertEqual(
                hc[p] + hr2[p, p],
                2 * one_body_integrals[p, p] + two_body_integrals[p, p, p, p])
            for q in range(2):
                if p != q:
                    self.assertEqual(hr1[p, q], two_body_integrals[p, p, q, q])
                    self.assertEqual(
                        hr2[p, q], 2 * two_body_integrals[p, q, q, p] -
                        two_body_integrals[p, q, p, q])


class DOCIHamiltonianTest(unittest.TestCase):

    def setUp(self):
        self.geometry = [('H', (0., 0., 0.)), ('H', (0., 0., 0.7414))]
        self.basis = 'sto-3g'
        self.multiplicity = 1
        self.filename = os.path.join(DATA_DIRECTORY, 'H2_sto-3g_singlet_0.7414')
        self.molecule = MolecularData(self.geometry,
                                      self.basis,
                                      self.multiplicity,
                                      filename=self.filename)
        self.molecule.load()

    def test_n_body_tensor_errors(self):
        doci_hamiltonian = DOCIHamiltonian.zero(n_qubits=2)
        with self.assertRaises(TypeError):
            doci_hamiltonian.n_body_tensors = 0
        with self.assertRaises(IndexError):
            _ = doci_hamiltonian[((0, 0), (0, 0))]
        with self.assertRaises(IndexError):
            _ = doci_hamiltonian[((0, 0), (0, 0), (0, 0), (0, 0))]
        with self.assertRaises(IndexError):
            _ = doci_hamiltonian[((1, 1), (0, 0))]
        with self.assertRaises(IndexError):
            _ = doci_hamiltonian[((0, 1), (2, 1), (3, 0), (8, 0))]

    def test_errors_operations(self):
        doci_hamiltonian = DOCIHamiltonian.zero(n_qubits=2)
        doci_hamiltonian2 = DOCIHamiltonian.zero(n_qubits=3)
        with self.assertRaises(TypeError):
            doci_hamiltonian += 'a'
        with self.assertRaises(TypeError):
            doci_hamiltonian -= 'a'
        with self.assertRaises(TypeError):
            doci_hamiltonian *= 'a'
        with self.assertRaises(TypeError):
            doci_hamiltonian /= 'a'
        with self.assertRaises(TypeError):
            doci_hamiltonian += doci_hamiltonian2
        with self.assertRaises(TypeError):
            doci_hamiltonian -= doci_hamiltonian2

    def test_adding_constants(self):
        doci_hamiltonian = DOCIHamiltonian.zero(n_qubits=2)
        doci_hamiltonian += 2
        self.assertAlmostEqual(doci_hamiltonian.constant, 2)
        doci_hamiltonian -= 3
        self.assertAlmostEqual(doci_hamiltonian.constant, -1)

    def test_basic_operations(self):
        doci_hamiltonian1 = DOCIHamiltonian.zero(n_qubits=2)
        doci_hamiltonian2 = DOCIHamiltonian.from_integrals(
            constant=self.molecule.nuclear_repulsion,
            one_body_integrals=self.molecule.one_body_integrals,
            two_body_integrals=self.molecule.two_body_integrals)
        self.assertTrue(doci_hamiltonian2 == doci_hamiltonian1 +
                        doci_hamiltonian2)
        self.assertTrue(doci_hamiltonian1 -
                        doci_hamiltonian2 == doci_hamiltonian2 / -1)
        self.assertTrue(doci_hamiltonian2 * 0 == doci_hamiltonian1)

    def test_error(self):
        doci_hamiltonian = DOCIHamiltonian.from_integrals(
            constant=self.molecule.nuclear_repulsion,
            one_body_integrals=self.molecule.one_body_integrals,
            two_body_integrals=self.molecule.two_body_integrals)
        with self.assertRaises(TypeError):
            doci_hamiltonian[((1, 0), (0, 1))] = 1
        with self.assertRaises(IndexError):
            _ = doci_hamiltonian[((1, 0),)]
        with self.assertRaises(IndexError):
            _ = doci_hamiltonian[((1, 1), (0, 0))]
        with self.assertRaises(IndexError):
            _ = doci_hamiltonian[((0, 1), (1, 1), (0, 0), (2, 0))]

    def test_getting_setting_constant(self):
        doci_hamiltonian = DOCIHamiltonian.zero(n_qubits=2)
        doci_hamiltonian.constant = 1
        self.assertEqual(doci_hamiltonian[()], 1)

    def test_getting_setting_1body(self):
        doci_hamiltonian = DOCIHamiltonian.zero(n_qubits=2)
        doci_hamiltonian.hc[0] = 2
        doci_hamiltonian.hc[1] = 4
        self.assertEqual(doci_hamiltonian[((0, 1), (0, 0))], 1)
        self.assertEqual(doci_hamiltonian[((1, 1), (1, 0))], 1)
        self.assertEqual(doci_hamiltonian[((2, 1), (2, 0))], 2)
        self.assertEqual(doci_hamiltonian[((3, 1), (3, 0))], 2)

    def test_getting_setting_hr2(self):
        doci_hamiltonian = DOCIHamiltonian.zero(n_qubits=2)
        doci_hamiltonian.hr2[0, 0] = 2
        doci_hamiltonian.hr2[1, 1] = 4
        self.assertEqual(doci_hamiltonian[((0, 1), (1, 1), (1, 0), (0, 0))], 1)
        self.assertEqual(doci_hamiltonian[((1, 1), (0, 1), (0, 0), (1, 0))], 1)
        self.assertEqual(doci_hamiltonian[((2, 1), (3, 1), (3, 0), (2, 0))], 2)
        self.assertEqual(doci_hamiltonian[((3, 1), (2, 1), (2, 0), (3, 0))], 2)

        doci_hamiltonian.hr2[0, 1] = 2
        self.assertEqual(doci_hamiltonian[((0, 1), (2, 1), (2, 0), (0, 0))], 1)
        self.assertEqual(doci_hamiltonian[((0, 1), (3, 1), (3, 0), (0, 0))], 1)
        self.assertEqual(doci_hamiltonian[((1, 1), (2, 1), (2, 0), (1, 0))], 1)
        self.assertEqual(doci_hamiltonian[((1, 1), (3, 1), (3, 0), (1, 0))], 1)

    def test_getting_setting_hr1(self):
        doci_hamiltonian = DOCIHamiltonian.zero(n_qubits=2)
        doci_hamiltonian.hr1[0, 1] = 2
        self.assertEqual(doci_hamiltonian[(0, 1), (1, 1), (3, 0), (2, 0)], 1)
        self.assertEqual(doci_hamiltonian[(1, 1), (0, 1), (2, 0), (3, 0)], 1)

    def test_from_integrals_to_qubit(self):
        hamiltonian = jordan_wigner(self.molecule.get_molecular_hamiltonian())
        doci_hamiltonian = DOCIHamiltonian.from_integrals(
            constant=self.molecule.nuclear_repulsion,
            one_body_integrals=self.molecule.one_body_integrals,
            two_body_integrals=self.molecule.two_body_integrals).qubit_operator

        hamiltonian_matrix = get_sparse_operator(hamiltonian).toarray()
        doci_hamiltonian_matrix = get_sparse_operator(
            doci_hamiltonian).toarray()
        diagonal = numpy.real(numpy.diag(hamiltonian_matrix))
        doci_diagonal = numpy.real(numpy.diag(doci_hamiltonian_matrix))
        position_of_doci_diag_in_h = [0] * len(doci_diagonal)
        for idx, doci_eigval in enumerate(doci_diagonal):
            closest_in_diagonal = None
            for idx2, eig in enumerate(diagonal):
                if closest_in_diagonal is None or abs(eig - doci_eigval) < abs(
                        closest_in_diagonal - doci_eigval):
                    closest_in_diagonal = eig
                    position_of_doci_diag_in_h[idx] = idx2
            assert abs(closest_in_diagonal - doci_eigval) < EQ_TOLERANCE, (
                "Value " + str(doci_eigval) + " of the DOCI Hamiltonian " +
                "diagonal did not appear in the diagonal of the full " +
                "Hamiltonian. The closest value was " +
                str(closest_in_diagonal))

        sub_matrix = hamiltonian_matrix[numpy.ix_(position_of_doci_diag_in_h,
                                                  position_of_doci_diag_in_h)]
        assert numpy.allclose(doci_hamiltonian_matrix, sub_matrix), (
            "The coupling between the DOCI states in the DOCI Hamiltonian " +
            "should be identical to that between these states in the full " +
            "Hamiltonian bur the DOCI hamiltonian matrix\n" +
            str(doci_hamiltonian_matrix) +
            "\ndoes not match the corresponding sub-matrix of the full " +
            "Hamiltonian\n" + str(sub_matrix))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
'''Tests for conversions.py'''
import unittest
import pytest
import numpy
import sympy

from openfermion.ops.operators import (QuadOperator, BosonOperator,
                                       FermionOperator, MajoranaOperator,
                                       QubitOperator)
from openfermion.transforms.repconversions.conversions import (
    get_diagonal_coulomb_hamiltonian)
from openfermion.transforms.opconversions.term_reordering import normal_ordered

from openfermion.transforms.opconversions.conversions import (
    get_quad_operator, get_boson_operator, get_majorana_operator,
    get_fermion_operator, check_no_sympy,
    _fermion_operator_to_majorana_operator, _fermion_term_to_majorana_operator)


class GetQuadOperatorTest(unittest.TestCase):

    def setUp(self):
        self.hbar = 0.5

    def test_invalid_op(self):
        op = QuadOperator()
        with self.assertRaises(TypeError):
            _ = get_quad_operator(op)

    def test_zero(self):
        b = BosonOperator()
        q = get_quad_operator(b)
        self.assertTrue(q == QuadOperator.zero())

    def test_identity(self):
        b = BosonOperator('')
        q = get_quad_operator(b)
        self.assertTrue(q == QuadOperator.identity())

    def test_creation(self):
        b = BosonOperator('0^')
        q = get_quad_operator(b, hbar=self.hbar)
        expected = QuadOperator('q0') - 1j * QuadOperator('p0')
        expected /= numpy.sqrt(2 * self.hbar)
        self.assertTrue(q == expected)

    def test_annihilation(self):
        b = BosonOperator('0')
        q = get_quad_operator(b, hbar=self.hbar)
        expected = QuadOperator('q0') + 1j * QuadOperator('p0')
        expected /= numpy.sqrt(2 * self.hbar)
        self.assertTrue(q == expected)

    def test_two_mode(self):
        b = BosonOperator('0^ 2')
        q = get_quad_operator(b, hbar=self.hbar)
        expected = QuadOperator('q0') - 1j * QuadOperator('p0')
        expected *= (QuadOperator('q2') + 1j * QuadOperator('p2'))
        expected /= 2 * self.hbar
        self.assertTrue(q == expected)

    def test_two_term(self):
        b = BosonOperator('0^ 0') + BosonOperator('0 0^')
        q = get_quad_operator(b, hbar=self.hbar)
        expected = (QuadOperator('q0') - 1j*QuadOperator('p0')) \
            * (QuadOperator('q0') + 1j*QuadOperator('p0')) \
            + (QuadOperator('q0') + 1j*QuadOperator('p0')) \
            * (QuadOperator('q0') - 1j*QuadOperator('p0'))
        expected /= 2 * self.hbar
        self.assertTrue(q == expected)

    def test_q_squared(self):
        b = self.hbar * (BosonOperator('0^ 0^') + BosonOperator('0 0') +
                         BosonOperator('') + 2 * BosonOperator('0^ 0')) / 2
        q = normal_ordered(get_quad_operator(b, hbar=self.hbar), hbar=self.hbar)
        expected = QuadOperator('q0 q0')
        self.assertTrue(q == expected)

    def test_p_squared(self):
        b = self.hbar * (-BosonOperator('1^ 1^') - BosonOperator('1 1') +
                         BosonOperator('') + 2 * BosonOperator('1^ 1')) / 2
        q = normal_ordered(get_quad_operator(b, hbar=self.hbar), hbar=self.hbar)
        expected = QuadOperator('p1 p1')
        self.assertTrue(q == expected)


class GetBosonOperatorTest(unittest.TestCase):

    def setUp(self):
        self.hbar = 0.5

    def test_invalid_op(self):
        op = BosonOperator()
        with self.assertRaises(TypeError):
            _ = get_boson_operator(op)

    def test_zero(self):
        q = QuadOperator()
        b = get_boson_operator(q)
        self.assertTrue(b == BosonOperator.zero())

    def test_identity(self):
        q = QuadOperator('')
        b = get_boson_operator(q)
        self.assertTrue(b == BosonOperator.identity())

    def test_x(self):
        q = QuadOperator('q0')
        b = get_boson_operator(q, hbar=self.hbar)
        expected = BosonOperator('0') + BosonOperator('0^')
        expected *= numpy.sqrt(self.hbar / 2)
        self.assertTrue(b == expected)

    def test_p(self):
        q = QuadOperator('p2')
        b = get_boson_operator(q, hbar=self.hbar)
        expected = BosonOperator('2') - BosonOperator('2^')
        expected *= -1j * numpy.sqrt(self.hbar / 2)
        self.assertTrue(b == expected)

    def test_two_mode(self):
        q = QuadOperator('p2 q0')
        b = get_boson_operator(q, hbar=self.hbar)
        expected = -1j*self.hbar/2 \
            * (BosonOperator('0') + BosonOperator('0^')) \
            * (BosonOperator('2') - BosonOperator('2^'))
        self.assertTrue(b == expected)

    def test_two_term(self):
        q = QuadOperator('p0 q0') + QuadOperator('q0 p0')
        b = get_boson_operator(q, hbar=self.hbar)
        expected = -1j*self.hbar/2 \
            * ((BosonOperator('0') + BosonOperator('0^'))
               * (BosonOperator('0') - BosonOperator('0^'))
               + (BosonOperator('0') - BosonOperator('0^'))
               * (BosonOperator('0') + BosonOperator('0^')))
        self.assertTrue(b == expected)


def test_get_fermion_operator_majorana_operator():
    a = MajoranaOperator((0, 3), 2.0) + MajoranaOperator((1, 2, 3))
    op = get_fermion_operator(a)
    expected_op = (-2j * (FermionOperator(((0, 0), (1, 0))) - FermionOperator(
        ((0, 0), (1, 1))) + FermionOperator(((0, 1), (1, 0))) - FermionOperator(
            ((0, 1), (1, 1)))) - 2 * FermionOperator(
                ((0, 0), (1, 1), (1, 0))) + 2 * FermionOperator(
                    ((0, 1), (1, 1), (1, 0))) + FermionOperator(
                        (0, 0)) - FermionOperator((0, 1)))
    assert normal_ordered(op) == normal_ordered(expected_op)


def test_get_fermion_operator_wrong_type():
    with pytest.raises(TypeError):
        _ = get_fermion_operator(QubitOperator())


class GetMajoranaOperatorTest(unittest.TestCase):
    """Test class get Majorana Operator."""

    def test_raises(self):
        """Test raises errors."""
        with self.assertRaises(TypeError):
            get_majorana_operator(1.0)
        with self.assertRaises(TypeError):
            _fermion_operator_to_majorana_operator([1.0])
        with self.assertRaises(TypeError):
            _fermion_term_to_majorana_operator(1.0)

    def test_get_majorana_operator_fermion_operator(self):
        """Test conversion FermionOperator to MajoranaOperator."""
        fermion_op = (-2j * (FermionOperator(
            ((0, 0), (1, 0))) - FermionOperator(
                ((0, 0), (1, 1))) + FermionOperator(
                    ((0, 1), (1, 0))) - FermionOperator(
                        ((0, 1), (1, 1)))) - 2 * FermionOperator(
                            ((0, 0), (1, 1), (1, 0))) + 2 * FermionOperator(
                                ((0, 1), (1, 1), (1, 0))) + FermionOperator(
                                    (0, 0)) - FermionOperator((0, 1)))

        majorana_op = get_majorana_operator(fermion_op)
        expected_op = (MajoranaOperator((0, 3), 2.0) + MajoranaOperator(
            (1, 2, 3)))
        self.assertTrue(majorana_op == expected_op)

    def test_get_majorana_operator_diagonalcoulomb(self):
        """Test get majorana from Diagonal Coulomb."""
        fermion_op = (FermionOperator('0^ 1', 1.0) +
                      FermionOperator('1^ 0', 1.0))

        diagonal_ham = get_diagonal_coulomb_hamiltonian(fermion_op)

        self.assertTrue(
            get_majorana_operator(diagonal_ham) == get_majorana_operator(
                fermion_op))


class RaisesSympyExceptionTest(unittest.TestCase):

    def test_raises_sympy_expression(self):
        operator = FermionOperator('0^', sympy.Symbol('x'))
        with self.assertRaises(TypeError):
            check_no_sympy(operator)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""FermionOperator stores a sum of products of fermionic ladder operators."""

from openfermion.ops.operators.symbolic_operator import SymbolicOperator


class FermionOperator(SymbolicOperator):
    r"""FermionOperator stores a sum of products of fermionic ladder operators.

    In OpenFermion, we describe fermionic ladder operators using the shorthand:
    'q^' = a^\dagger_q
    'q' = a_q
    where {'p^', 'q'} = delta_pq

    One can multiply together these fermionic ladder operators to obtain a
    fermionic term. For instance, '2^ 1' is a fermion term which
    creates at orbital 2 and destroys at orbital 1. The FermionOperator class
    also stores a coefficient for the term, e.g. '3.17 * 2^ 1'.

    The FermionOperator class is designed (in general) to store sums of these
    terms. For instance, an instance of FermionOperator might represent
    3.17 2^ 1 - 66.2 * 8^ 7 6^ 2
    The Fermion Operator class overloads operations for manipulation of
    these objects by the user.

    FermionOperator is a subclass of SymbolicOperator. Importantly, it has
    attributes set as follows::

        actions = (1, 0)
        action_strings = ('^', '')
        action_before_index = False
        different_indices_commute = False

    See the documentation of SymbolicOperator for more details.

    Example:
        .. code-block:: python

            ham = (FermionOperator('0^ 3', .5)
                   + .5 * FermionOperator('3^ 0'))
            # Equivalently
            ham2 = FermionOperator('0^ 3', 0.5)
            ham2 += FermionOperator('3^ 0', 0.5)

    Note:
        Adding FermionOperators is faster using += (as this
        is done by in-place addition). Specifying the coefficient
        during initialization is faster than multiplying a FermionOperator
        with a scalar.
    """

    @property
    def actions(self):
        """The allowed actions."""
        return (1, 0)

    @property
    def action_strings(self):
        """The string representations of the allowed actions."""
        return ('^', '')

    @property
    def action_before_index(self):
        """Whether action comes before index in string representations."""
        return False

    @property
    def different_indices_commute(self):
        """Whether factors acting on different indices commute."""
        return False

    def is_normal_ordered(self):
        """Return whether or not term is in normal order.

        In our convention, normal ordering implies terms are ordered
        from highest tensor factor (on left) to lowest (on right). Also,
        ladder operators come first.
        """
        for term in self.terms:
            for i in range(1, len(term)):
                for j in range(i, 0, -1):
                    right_operator = term[j]
                    left_operator = term[j - 1]
                    if right_operator[1] and not left_operator[1]:
                        return False
                    elif (right_operator[1] == left_operator[1] and
                          right_operator[0] >= left_operator[0]):
                        return False
        return True

    def is_two_body_number_conserving(self, check_spin_symmetry=False):
        """Query whether operator has correct form to be from a molecule.

        Require that term is particle-number conserving (same number of
        raising and lowering operators). Require that term has 0, 2 or 4
        ladder operators. Require that term conserves spin (parity of
        raising operators equals parity of lowering operators).

        Args:
            check_spin_symmetry (bool): Whether to check if
                operator conserves spin.
        """
        for term in self.terms:
            if len(term) not in (0, 2, 4):
                return False

            # Make sure term conserves particle number and (optionally) spin.
            spin = 0
            particles = 0
            for operator in term:
                particles += (-1)**operator[1]  # add 1 if create, else -1
                spin += (-1)**(operator[0] + operator[1])
            if particles:
                return False
            elif spin and check_spin_symmetry:
                return False
        return True



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for _qubit_partitioning.py"""
import unittest
import numpy
from .fermion_partitioning import (pair_within, pair_within_simultaneously,
                                   pair_within_simultaneously_binned,
                                   pair_within_simultaneously_symmetric,
                                   _gen_partitions, _get_padding,
                                   _parallel_iter, _asynchronous_iter)


class TestPairWithin(unittest.TestCase):

    def test_zero(self):
        count = 0
        for _ in pair_within([]):
            count += 1  # pragma: no cover
        self.assertEqual(count, 0)

    def test_pair(self):
        labels = [0, 1]
        pairings_list = [pairing for pairing in pair_within(labels)]
        self.assertEqual(len(pairings_list), 1)
        self.assertEqual(len(pairings_list[0]), 1)
        self.assertEqual(pairings_list[0][0], (0, 1))

    def test_threeway(self):
        labels = [0, 1, 2]
        pairings_list = [pairing for pairing in pair_within(labels)]
        self.assertEqual(len(pairings_list), 3)
        all_pairs = [(0, 1), (0, 2), (1, 2)]
        checksum = [0, 0, 0]
        for pairing in pairings_list:
            self.assertEqual(len(pairing), 2)
            for j in range(3):
                if all_pairs[j] in pairing:
                    checksum[j] += 1
        print(checksum)
        for j in range(3):
            self.assertEqual(checksum[j], 1)

    def test_many(self):
        for num_indices in range(2, 16):
            print()
            print(num_indices)
            labels = [j for j in range(num_indices)]
            pairings_list = [pairing for pairing in pair_within(labels)]
            if num_indices % 2 == 0:
                self.assertEqual(len(pairings_list), num_indices - 1)
            else:
                self.assertEqual(len(pairings_list), num_indices)
            for pairing in pairings_list:
                print(pairing)
            all_pairs = [(i, j)
                         for i in range(num_indices)
                         for j in range(i + 1, num_indices)]
            checksum = [0 for _ in all_pairs]
            for pairing in pairings_list:
                for j, pair in enumerate(all_pairs):
                    if pair in pairing:
                        checksum[j] += 1
            print(checksum)
            for cs in checksum:
                self.assertGreaterEqual(cs, 1)


class TestPairWithinSimultaneously(unittest.TestCase):

    def test_small(self):
        for num_indices in [4, 5, 7, 10, 15]:
            print()
            print('Trying with num_indices = {}'.format(num_indices))
            labels = [j for j in range(num_indices)]
            all_quads = [((i, j, k, l)) for i in range(num_indices)
                         for j in range(i + 1, num_indices)
                         for k in range(j + 1, num_indices)
                         for l in range(k + 1, num_indices)]
            checksum = [0 for pp2 in all_quads]
            for pairing in pair_within_simultaneously(labels):
                print(pairing)
                for j, quad in enumerate(all_quads):
                    for n1 in range(3):
                        n2 = (n1 + 1) % 3
                        n3 = (n1 + 2) % 3
                        pair1 = (quad[0], quad[n1 + 1])
                        pair2 = (quad[min(n2, n3) + 1], quad[max(n2, n3) + 1])
                        if pair1 in pairing and pair2 in pairing:
                            checksum[j] += 1

            for j in range(len(checksum)):
                print(all_quads[j], checksum[j])
                self.assertGreaterEqual(checksum[j], 1)


class TestPadding(unittest.TestCase):

    def test_primes(self):
        bin_size = 11
        for num_bins in range(10):
            self.assertEqual(_get_padding(num_bins, bin_size), bin_size)

    def test_composite(self):
        bin_size = 12
        for num_bins in range(4, 13):
            self.assertEqual(_get_padding(num_bins, bin_size), 13)


class TestIterators(unittest.TestCase):

    def test_asynchronous_three(self):
        lists = [[0, 1, 2], [0, 1, 2], [0, 1, 2], [0, 1, 2]]
        test_matrices = [
            [numpy.zeros((3, 3)) for j in range(4)] for k in range(4)
        ]
        iterator = _asynchronous_iter(lists)
        count = 0
        for next_tuple in iterator:
            print(next_tuple)
            count += 1
            for j in range(4):
                for k in range(4):
                    test_matrices[j][k][next_tuple[j], next_tuple[k]] += 1

        self.assertEqual(count, 9)
        for j in range(4):
            for k in range(j + 1, 4):
                for i1 in range(3):
                    for i2 in range(3):
                        self.assertEqual(test_matrices[j][k][i1, i2], 1)

    def test_asynchronous_four(self):
        lists = [[x for x in range(4)] for y in range(4)]
        iterator = _asynchronous_iter(lists)
        count = 0
        for _ in iterator:
            count += 1
        self.assertEqual(count, 25)

    def test_parallel(self):

        def iter1():
            for j in range(4):
                yield j

        def iter2():
            for j in range(5):
                yield j

        iterators = [iter1(), iter2()]
        count = 0
        for _ in _parallel_iter(iterators):
            count += 1
        self.assertEqual(count, 5)

    def test_parallel_flatten(self):

        def iter1():
            for j in range(4):
                yield [j]

        def iter2():
            for j in range(5):
                yield [j]

        iterators = [iter1(), iter2()]
        for res in _parallel_iter(iterators, flatten=False):
            self.assertTrue(isinstance(res[0], list))
        iterators = [iter1(), iter2()]
        for res in _parallel_iter(iterators, flatten=True):
            self.assertTrue(isinstance(res[0], int))


class TestPairingWithSymmetries(unittest.TestCase):

    def test_two_fermions(self):
        bins = [[1], [2], [3], [4]]
        count = 0
        for pairing in pair_within_simultaneously_binned(bins):
            count += 1
            self.assertEqual(len(pairing), 2)
            print(pairing)
        self.assertEqual(count, 1)
        count = 0
        for pairing in pair_within_simultaneously_symmetric(2, 2):
            count += 1
            self.assertEqual(len(pairing), 2)
        self.assertEqual(count, 1)
        count = 0
        for pairing in pair_within_simultaneously_symmetric(2, 1):
            count += 1
            self.assertEqual(len(pairing), 2)
        self.assertEqual(count, 1)
        count = 0
        for pairing in pair_within_simultaneously_symmetric(2, 0):
            count += 1
            self.assertEqual(len(pairing), 2)
        self.assertEqual(count, 1)

    def test_four_fermions(self):
        print('Trying with 0 symmetries')
        count = 0
        for pairing in pair_within_simultaneously_symmetric(4, 0):
            print(pairing)
            count += 1
            self.assertEqual(len(pairing), 4)
        self.assertEqual(count, 18)
        print('Trying with 1 symmetry')
        count = 0
        for pairing in pair_within_simultaneously_symmetric(4, 1):
            print(pairing)
            count += 1
            self.assertEqual(len(pairing), 4)
        self.assertEqual(count, 10)
        print('Trying with 2 symmetries')
        count = 0
        for pairing in pair_within_simultaneously_symmetric(4, 2):
            print(pairing)
            count += 1
            self.assertEqual(len(pairing), 4)
        self.assertEqual(count, 5)
        print('Trying with 3 symmetries')
        count = 0
        for pairing in pair_within_simultaneously_symmetric(4, 3):
            print(pairing)
            count += 1
            self.assertEqual(len(pairing), 4)
        self.assertEqual(count, 3)

    def test_four_symmetries(self):
        for num_fermions in [5, 8, 9]:
            for _ in pair_within_simultaneously_symmetric(num_fermions, 3):
                pass


def test_gen_partitions_1input():
    labels = [
        0,
    ]
    count = 0
    for _ in _gen_partitions(labels):
        count += 1
    assert count == 1



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import cirq

from openfermion import swap_network


def test_swap_network():
    n_qubits = 4
    qubits = cirq.LineQubit.range(n_qubits)

    circuit = cirq.Circuit(
        swap_network(qubits, lambda i, j, q0, q1: cirq.ISWAP(q0, q1)**-1))
    cirq.testing.assert_has_diagram(
        circuit, """
0: ───iSwap──────×──────────────────iSwap──────×──────────────────
      │          │                  │          │
1: ───iSwap^-1───×───iSwap──────×───iSwap^-1───×───iSwap──────×───
                     │          │                  │          │
2: ───iSwap──────×───iSwap^-1───×───iSwap──────×───iSwap^-1───×───
      │          │                  │          │
3: ───iSwap^-1───×──────────────────iSwap^-1───×──────────────────
""")

    circuit = cirq.Circuit(
        swap_network(qubits,
                     lambda i, j, q0, q1: cirq.ISWAP(q0, q1)**-1,
                     fermionic=True,
                     offset=True))
    cirq.testing.assert_has_diagram(circuit,
                                    """
0     1        2        3
│     │        │        │
│     iSwap────iSwap^-1 │
│     │        │        │
│     ×ᶠ───────×ᶠ       │
│     │        │        │
iSwap─iSwap^-1 iSwap────iSwap^-1
│     │        │        │
×ᶠ────×ᶠ       ×ᶠ───────×ᶠ
│     │        │        │
│     iSwap────iSwap^-1 │
│     │        │        │
│     ×ᶠ───────×ᶠ       │
│     │        │        │
iSwap─iSwap^-1 iSwap────iSwap^-1
│     │        │        │
×ᶠ────×ᶠ       ×ᶠ───────×ᶠ
│     │        │        │
""",
                                    transpose=True)

    n_qubits = 5
    qubits = cirq.LineQubit.range(n_qubits)

    circuit = cirq.Circuit(swap_network(
        qubits, lambda i, j, q0, q1: (), fermionic=True),
                           strategy=cirq.InsertStrategy.EARLIEST)
    cirq.testing.assert_has_diagram(
        circuit, """
0: ───×ᶠ────────×ᶠ────────×ᶠ───
      │         │         │
1: ───×ᶠ───×ᶠ───×ᶠ───×ᶠ───×ᶠ───
           │         │
2: ───×ᶠ───×ᶠ───×ᶠ───×ᶠ───×ᶠ───
      │         │         │
3: ───×ᶠ───×ᶠ───×ᶠ───×ᶠ───×ᶠ───
           │         │
4: ────────×ᶠ────────×ᶠ────────
""")

    circuit = cirq.Circuit(swap_network(
        qubits, lambda i, j, q0, q1: (), offset=True),
                           strategy=cirq.InsertStrategy.EARLIEST)
    cirq.testing.assert_has_diagram(circuit,
                                    """
0 1 2 3 4
│ │ │ │ │
│ ×─× ×─×
│ │ │ │ │
×─× ×─× │
│ │ │ │ │
│ ×─× ×─×
│ │ │ │ │
×─× ×─× │
│ │ │ │ │
│ ×─× ×─×
│ │ │ │ │
""",
                                    transpose=True)


def test_reusable():
    ops = swap_network(cirq.LineQubit.range(5))
    assert list(ops) == list(ops)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""The MajoranaOperator data structure."""

import itertools
import numpy


class MajoranaOperator:
    r"""A linear combination of products of Majorana operators.

    A system of N fermionic modes can be described using 2N Majorana operators
    :math:`\gamma_1, \ldots, \gamma_{2N}`
    as an alternative to using N fermionic annihilation operators. The algebra
    of Majorana operators amounts to the relation

    .. math::
        \{\gamma_i, \gamma_j\}
         = \gamma_i \gamma_j + \gamma_j \gamma_i
         = 2 \delta_{ij}

    Note that this implies :math:`\gamma_i^2 = 1`.

    The MajoranaOperator class stores a linear combination of products
    of Majorana operators. Each product is represented as a tuple of
    integers representing the indices of the operators. As an example,
    `MajoranaOperator((2, 3, 5), -1.5)` initializes an operator with
    a single term which represents the operator
    :math:`-1.5 \gamma_2 \gamma_3 \gamma_5`. MajoranaOperators can be
    added, subtracted, multiplied, and divided by scalars. They can be
    compared for approximate numerical equality using `==`.

    Attributes:
        terms: A dictionary from term, represented by a tuple of integers,
        to the coefficient of the term in the linear combination.
    """

    def __init__(self, term=None, coefficient=1.0):
        """Initialize a MajoranaOperator with a single term.

        Args:
            term (Tuple[int]): The indices of a Majorana operator term
                to start off with
            coefficient (complex): The coefficient of the term

        Returns:
            MajoranaOperator
        """
        self.terms = {}
        if term is not None:
            term, parity = _sort_majorana_term(term)
            self.terms[term] = coefficient * (-1)**parity

    @staticmethod
    def from_dict(terms):
        """Initialize a MajoranaOperator from a terms dictionary.

        WARNING: The given dictionary is not validated whatsoever. It's up
        to you to ensure that it is properly formed.

        Args:
            terms: A dictionary from Majorana term to coefficient
        """
        op = MajoranaOperator()
        op.terms = terms
        return op

    def commutes_with(self, other):
        """Test commutation with another MajoranaOperator"""
        if not isinstance(other, type(self)):
            raise TypeError(
                'Can only test commutation with another MajoranaOperator.')

        if len(self.terms) == 1 and len(other.terms) == 1:
            return _majorana_terms_commute(
                list(self.terms.keys())[0],
                list(other.terms.keys())[0])
        return self * other == other * self

    def with_basis_rotated_by(self, transformation_matrix):
        r"""Change to a basis of new Majorana operators.

        The input to this method is a real orthogonal matrix :math:`O`.
        It returns a new MajoranaOperator which is equivalent to the old one
        but rewritten in terms of a new basis of Majorana operators.
        Let the original Majorana operators be denoted by
        :math:`\gamma_i` and the new operators be denoted by
        :math:`\tilde{\gamma_i}`. Then they are related by the equation

        .. math::

            \tilde{\gamma_i} = \sum_j O_{ij} \gamma_j.

        Args:
            transformation_matrix: A real orthogonal matrix representing
                the basis transformation.
        Returns:
            The rotated operator.
        """
        if not _is_real_orthogonal(transformation_matrix):
            raise ValueError("Transformation matrix is not real orthogonal.")

        rotated_op = MajoranaOperator()
        for term, coeff in self.terms.items():
            rotated_term = _rotate_basis(term, transformation_matrix)
            rotated_term *= coeff
            rotated_op += rotated_term
        return rotated_op

    def __iadd__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        for term, coefficient in other.terms.items():
            if term in self.terms:
                self.terms[term] += coefficient
            else:
                self.terms[term] = coefficient

        return self

    def __add__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        terms = {}
        terms.update(self.terms)

        for term, coefficient in other.terms.items():
            if term in terms:
                terms[term] += coefficient
            else:
                terms[term] = coefficient

        return MajoranaOperator.from_dict(terms)

    def __isub__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        for term, coefficient in other.terms.items():
            if term in self.terms:
                self.terms[term] -= coefficient
            else:
                self.terms[term] = coefficient

        return self

    def __sub__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        terms = {}
        terms.update(self.terms)
        for term, coefficient in other.terms.items():
            if term in terms:
                terms[term] -= coefficient
            else:
                terms[term] = -coefficient
        return MajoranaOperator.from_dict(terms)

    def __mul__(self, other):
        if not isinstance(other, (type(self), int, float, complex)):
            return NotImplemented

        if isinstance(other, (int, float, complex)):
            terms = {
                term: coefficient * other
                for term, coefficient in self.terms.items()
            }
            return MajoranaOperator.from_dict(terms)

        terms = {}
        for left_term, left_coefficient in self.terms.items():
            for right_term, right_coefficient in other.terms.items():
                new_term, parity = _merge_majorana_terms(left_term, right_term)
                coefficient = left_coefficient * right_coefficient * (
                    -1)**parity
                if new_term in terms:
                    terms[new_term] += coefficient
                else:
                    terms[new_term] = coefficient
        return MajoranaOperator.from_dict(terms)

    def __imul__(self, other):
        if not isinstance(other, (type(self), int, float, complex)):
            return NotImplemented

        if isinstance(other, (int, float, complex)):
            for term in self.terms:
                self.terms[term] *= other
            return self

        return self * other

    def __rmul__(self, other):
        if not isinstance(other, (int, float, complex)):
            return NotImplemented
        return self * other

    def __truediv__(self, other):
        if not isinstance(other, (int, float, complex)):
            return NotImplemented

        terms = {
            term: coefficient / other
            for term, coefficient in self.terms.items()
        }
        return MajoranaOperator.from_dict(terms)

    def __itruediv__(self, other):
        if not isinstance(other, (int, float, complex)):
            return NotImplemented

        for term in self.terms:
            self.terms[term] /= other
        return self

    def __pow__(self, other):
        if not isinstance(other, int):
            return NotImplemented

        if other < 0:
            raise TypeError('Cannot raise MajoranaOperator to negative power.')

        result = MajoranaOperator(())
        for _ in range(other):
            result *= self
        return result

    def __neg__(self):
        return -1 * self

    def __eq__(self, other):
        """Approximate numerical equality."""
        if not isinstance(other, type(self)):
            return NotImplemented

        for term in self.terms.keys() | other.terms.keys():
            if term in self.terms and term in other.terms:
                if not numpy.isclose(self.terms[term], other.terms[term]):
                    return False
            elif term in self.terms:
                if not numpy.isclose(self.terms[term], 0.0):
                    return False
            else:
                if not numpy.isclose(other.terms[term], 0.0):
                    return False
        return True

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        if not self.terms:
            return '0'
        lines = []
        for term, coeff in sorted(self.terms.items()):
            if numpy.isclose(coeff, 0.0):
                continue
            lines.append('{} {} +'.format(coeff, term))
        if not lines:
            return '0'
        return '\n'.join(lines)[:-2]

    def __repr__(self):
        return 'MajoranaOperator.from_dict(terms={!r})'.format(self.terms)


def _sort_majorana_term(term):
    """Sort a Majorana term.

    Args:
        term (Tuple[int]): The indices of a Majorana operator term

    Returns:
        Tuple[Tuple[int], int]. The first object returned is a sorted list
        representing the indices acted upon. The second object is the parity
        of the term. A parity of 1 indicates that the term should include
        a minus sign.
    """
    if len(term) < 2:
        return term, 0
    center = len(term) // 2
    left_term, left_parity = _sort_majorana_term(term[:center])
    right_term, right_parity = _sort_majorana_term(term[center:])
    merged_term, merge_parity = _merge_majorana_terms(left_term, right_term)
    return merged_term, (left_parity + right_parity + merge_parity) % 2


def _merge_majorana_terms(left_term, right_term):
    """Merge two Majorana terms.

    Args:
        left_term (Tuple[int]): The left-hand term
        right_term (Tuple[int]): The right-hand term

    Returns:
        Tuple[Tuple[int], int]. The first object returned is a sorted list
        representing the indices acted upon. The second object is the parity
        of the term. A parity of 1 indicates that the term should include
        a minus sign.
    """
    merged_term = []
    parity = 0
    i, j = 0, 0
    while i < len(left_term) and j < len(right_term):
        if left_term[i] < right_term[j]:
            merged_term.append(left_term[i])
            i += 1
        elif left_term[i] > right_term[j]:
            merged_term.append(right_term[j])
            j += 1
            parity += len(left_term) - i
        else:
            parity += len(left_term) - i - 1
            i += 1
            j += 1
    if i == len(left_term):
        merged_term.extend(right_term[j:])
    else:
        merged_term.extend(left_term[i:])
    return tuple(merged_term), parity % 2


def _majorana_terms_commute(term_a, term_b):
    """Whether two Majorana terms commute.

    Args:
        term_a (Tuple[int]): The indices of a Majorana operator term
        term_b (Tuple[int]): The indices of a Majorana operator term

    Returns:
        bool. Whether The terms commute.
    """
    intersection = 0
    i, j = 0, 0
    while i < len(term_a) and j < len(term_b):
        if term_a[i] < term_b[j]:
            i += 1
        elif term_a[i] > term_b[j]:
            j += 1
        else:
            intersection += 1
            i += 1
            j += 1
    parity = (len(term_a) * len(term_b) - intersection) % 2
    return not parity


def _rotate_basis(term, transformation_matrix):
    n = transformation_matrix.shape[0]
    rotated_op = MajoranaOperator()
    for tup in itertools.product(range(n), repeat=len(term)):
        coeff = 1.0
        for i, j in zip(term, tup):
            coeff *= transformation_matrix[j, i]
        rotated_op += MajoranaOperator(tup, coeff)
    return rotated_op


def _is_real_orthogonal(matrix):
    n, m = matrix.shape
    return (n == m and numpy.allclose(numpy.imag(matrix), 0.0) and
            numpy.allclose(numpy.dot(matrix.T, matrix), numpy.eye(n)))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
A routine for constructing a circuit to exactly implement a unitary generated by
one-body rotations through the optimal Givens rotation network.  Construction
of this circuit can be found in Optica Vol. 3, Issue 12, pp. 1460-1465 (2016).
This Givens network improves upon the parallel Givens network for implementing
basis rotations in Phys. Rev. Lett. 120, 110501 (2018).
"""
from typing import cast, Iterable, Sequence, Tuple

import numpy
import cirq

from openfermion.linalg import givens_matrix_elements, givens_rotate
from openfermion.circuits.gates import Ryxxy


class GivensTranspositionError(Exception):
    pass


class GivensMatrixError(Exception):
    pass


def optimal_givens_decomposition(qubits: Sequence[cirq.Qid],
                                 unitary: numpy.ndarray
                                ) -> Iterable[cirq.Operation]:
    r"""
    Implement a circuit that provides the unitary that is generated by
    single-particle fermion generators

    .. math::

        U(v) = exp(log(v)_{p,q}(a_{p}^{\dagger}a_{q} - a_{q}^{\dagger}a_{p})

    This can be used for implementing an exact single-body basis rotation

    Args:
        qubits: Sequence of qubits to apply the operations over.  The qubits
                should be ordered in linear physical order.
        unitary:
    """
    N = unitary.shape[0]
    right_rotations = []
    left_rotations = []
    for i in range(1, N):
        if i % 2 == 1:
            for j in range(0, i):
                # eliminate U[N - j, i - j] by mixing U[N - j, i - j],
                # U[N - j, i - j - 1] by right multiplication
                # of a givens rotation matrix in column [i - j, i - j + 1]
                gmat = givens_matrix_elements(unitary[N - j - 1, i - j - 1],
                                              unitary[N - j - 1, i - j - 1 + 1],
                                              which='left')
                right_rotations.append((gmat.T, (i - j - 1, i - j)))
                givens_rotate(unitary,
                              gmat.conj(),
                              i - j - 1,
                              i - j,
                              which='col')
        else:
            for j in range(1, i + 1):
                # elimination of U[N + j - i, j] by mixing U[N + j - i, j] and
                # U[N + j - i - 1, j] by left multiplication
                # of a givens rotation that rotates row space
                # [N + j - i - 1, N + j - i
                gmat = givens_matrix_elements(unitary[N + j - i - 1 - 1, j - 1],
                                              unitary[N + j - i - 1, j - 1],
                                              which='right')
                left_rotations.append((gmat, (N + j - i - 2, N + j - i - 1)))
                givens_rotate(unitary,
                              gmat,
                              N + j - i - 2,
                              N + j - i - 1,
                              which='row')

    new_left_rotations = []
    for (left_gmat, (i, j)) in reversed(left_rotations):
        phase_matrix = numpy.diag([unitary[i, i], unitary[j, j]])
        matrix_to_decompose = left_gmat.conj().T.dot(phase_matrix)
        new_givens_matrix = givens_matrix_elements(matrix_to_decompose[1, 0],
                                                   matrix_to_decompose[1, 1],
                                                   which='left')
        new_phase_matrix = matrix_to_decompose.dot(new_givens_matrix.T)

        # check if T_{m,n}^{-1}D  = D T.
        # coverage: ignore
        if not numpy.allclose(new_phase_matrix.dot(new_givens_matrix.conj()),
                              matrix_to_decompose):
            raise GivensTranspositionError("Failed to shift the phase matrix "
                                           "from right to left")
        # coverage: ignore

        unitary[i, i], unitary[j, j] = new_phase_matrix[0, 0], new_phase_matrix[
            1, 1]
        new_left_rotations.append((new_givens_matrix.conj(), (i, j)))

    phases = numpy.diag(unitary)
    rotations = []
    ordered_rotations = []
    for (gmat, (i, j)) in list(reversed(new_left_rotations)) + list(
            map(lambda x: (x[0].conj().T, x[1]), reversed(right_rotations))):
        ordered_rotations.append((gmat, (i, j)))

        # if this throws the impossible has happened
        # coverage: ignore
        if not numpy.isclose(gmat[0, 0].imag, 0.0):
            raise GivensMatrixError(
                "Givens matrix does not obey our convention that all elements "
                "in the first column are real")
        if not numpy.isclose(gmat[1, 0].imag, 0.0):
            raise GivensMatrixError(
                "Givens matrix does not obey our convention that all elements "
                "in the first column are real")
        # coverage: ignore

        theta = numpy.arcsin(numpy.real(gmat[1, 0]))
        phi = numpy.angle(gmat[1, 1])
        rotations.append((i, j, theta, phi))

    for op in reversed(rotations):
        i, j, theta, phi = cast(Tuple[int, int, float, float], op)
        if not numpy.isclose(phi, 0.0):
            yield cirq.Z(qubits[j])**(phi / numpy.pi)

        yield Ryxxy(-theta).on(qubits[i], qubits[j])

    for idx, phase in enumerate(phases):
        yield cirq.Z(qubits[idx])**(numpy.angle(phase) / numpy.pi)



from .conversions import (
    get_interaction_operator,
    get_diagonal_coulomb_hamiltonian,
    get_molecular_data,
    get_quadratic_hamiltonian,
)

from .fourier_transforms import (
    fourier_transform,
    inverse_fourier_transform,
)

from .operator_tapering import (
    freeze_orbitals,
    prune_unused_indices,
)

from .qubit_operator_transforms import (
    project_onto_sector,
    projection_error,
    rotate_qubit_by_pauli,
)

from .qubit_tapering_from_stabilizer import (
    StabilizerError,
    check_commuting_stabilizers,
    check_stabilizer_linearity,
    reduce_number_of_terms,
    taper_off_qubits,
    fix_single_term,
)

from .weyl_ordering import (
    mccoy,
    weyl_polynomial_quantization,
    symmetric_ordering,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import unittest

from openfermion.ops.operators import BinaryPolynomial
from openfermion.ops.operators.binary_code import (BinaryCode, BinaryCodeError,
                                                   shift_decoder)


class CodeOperatorTest(unittest.TestCase):

    def test_init_errors(self):
        with self.assertRaises(TypeError):
            BinaryCode(
                1, [BinaryPolynomial(' w1 + w0 '),
                    BinaryPolynomial('w0 + 1')])
        with self.assertRaises(TypeError):
            BinaryCode([[0, 1], [1, 0]], '1+w1')
        with self.assertRaises(BinaryCodeError):
            BinaryCode([[0, 1], [1, 0]], [BinaryPolynomial(' w1 + w0 ')])
        with self.assertRaises(TypeError):
            BinaryCode([[0, 1, 1], [1, 0, 0]],
                       ['1 + w1', BinaryPolynomial('1 + w0'), 2.0])
        with self.assertRaises(BinaryCodeError):
            BinaryCode([[0, 1], [1, 0]],
                       [BinaryPolynomial(' w0 '),
                        BinaryPolynomial('w0 + 1')])
        with self.assertRaises(BinaryCodeError):
            BinaryCode([[0, 1], [1, 0]],
                       [BinaryPolynomial(' w5 '),
                        BinaryPolynomial('w0 + 1')])

    def test_addition(self):
        a = BinaryCode([[0, 1, 0], [1, 0, 1]], [
            BinaryPolynomial(' w1 + w0 '),
            BinaryPolynomial('w0 + 1'),
            BinaryPolynomial('w1')
        ])
        d = BinaryCode([[0, 1], [1, 0]],
                       [BinaryPolynomial(' w0 '),
                        BinaryPolynomial('w0 w1')])
        summation = a + d
        self.assertEqual(
            str(summation), "[[[0, 1, 0, 0, 0],"
            " [1, 0, 1, 0, 0],"
            " [0, 0, 0, 0, 1],"
            " [0, 0, 0, 1, 0]],"
            " '[[W1] + [W0],[W0] + [1],"
            "[W1],[W2],[W2 W3]]']")
        with self.assertRaises(TypeError):
            summation += 5

    def test_multiplication(self):
        a = BinaryCode([[0, 1, 0], [1, 0, 1]], [
            BinaryPolynomial(' w1 + w0 '),
            BinaryPolynomial('w0 + 1'),
            BinaryPolynomial('w1')
        ])
        d = BinaryCode([[0, 1], [1, 0]],
                       [BinaryPolynomial(' w0 '),
                        BinaryPolynomial('w0 w1')])

        b = a * d
        self.assertEqual(
            b.__repr__(), "[[[1, 0, 1], [0, 1, 0]],"
            " '[[W0] + [W0 W1],[1] +"
            " [W0],[W0 W1]]']")
        b = 2 * d
        self.assertEqual(
            str(b), "[[[0, 1, 0, 0], [1, 0, 0, 0], "
            "[0, 0, 0, 1], [0, 0, 1, 0]], "
            "'[[W0],[W0 W1],[W2],[W2 W3]]']")
        with self.assertRaises(BinaryCodeError):
            _ = d * a
        with self.assertRaises(TypeError):
            _ = 2.0 * a
        with self.assertRaises(TypeError):
            a *= 2.0
        with self.assertRaises(ValueError):
            b = 0 * a

    def test_shift(self):
        decoder = [BinaryPolynomial('1'), BinaryPolynomial('1 + w1 w0')]
        with self.assertRaises(TypeError):
            shift_decoder(decoder, 2.5)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tools to reduce the number of terms and taper off qubits
using stabilizer conditions. Based on ideas of arXiv:1701.08213. """

import numpy
from openfermion.ops.operators import QubitOperator
from openfermion.config import EQ_TOLERANCE
from openfermion.utils import operator_utils as op_utils


class StabilizerError(Exception):
    """Stabilizer error class."""

    def __init__(self, message):
        """
        Throw custom errors connected to stabilizers.

        Args:
            message(str): custome error message string.
        """
        Exception.__init__(self, message)


def check_commuting_stabilizers(stabilizer_list, msg, thres=EQ_TOLERANCE):
    """
    Auxiliary function checking that stabilizers commute.

    If two stabilizers anti-commute their product
    will have an imaginary coefficient.
    This function checks the list of stabilizers (QubitOperator)
    and raises and error if a complex number is found in
    any of the coefficients.

    Args:
        stabilizer_list (list): List of stabilizers as QubitOperators.
        msg (str): Message for the error.
        thres: Tolerance value, set to OpenFermion tolerance by default.
    """
    for stab in stabilizer_list:
        if abs(numpy.imag(list(stab.terms.values())[0])) >= thres:
            raise StabilizerError(msg)


def check_stabilizer_linearity(stabilizer_list, msg):
    """
    Auxiliary function checking that stabilizers are linearly independent.

    If two stabilizer are linearly dependent the result
    after some of their products will be the identity.
    This function checks the list of stabilizers (QubitOperator)
    and raises an error if the identity is found.

    Args:
        stabilizer_list (list): List of stabilizers (QubitOperator).
        msg (str): Message for the error.
    """
    for stab in stabilizer_list:
        if list(stab.terms.keys())[0] == ():
            raise StabilizerError(msg)


def fix_single_term(term, position, fixed_op, other_op, stabilizer):
    """
    Auxiliary function for term reductions.

    Automatically multiplies a single term with a given stabilizer if
    the Pauli operator on a given qubit is of one of two specified types.
    This fixes a certain representation of a logical operator.

    Args:
        term (QubitOperator): Single term to fix.
        position (int): Index of the qubit which is to be fixed.
        fixed_op (str): Pauli operator, which
                        will cause a multiplication by the
                        stabilizer when encountered at the fixed
                        position.
        other_op (str): Alternative Pauli operator, which will also
                        cause the multiplication by the stabilizer.
        stabilizer (QubitOperator): Stabilizer that is multiplied
                                    when necessary.

    Returns:
        term (QubitOperator): Updated term in a fiixed representation.
    """
    pauli_tuple = list(term.terms)[0]
    if (position, fixed_op) in pauli_tuple or (position,
                                               other_op) in pauli_tuple:
        return term * stabilizer
    else:
        return term


def _lookup_term(pauli_string, updated_terms_1, updated_terms_2):
    """
    Auxiliary function for reducing terms keeping length.

    This function checks the length of the original Pauli strings,
    compares it to a list of strings and returns the shortest operator
    equivalent to the original.

    Args:
        pauli_string (tuple): Original Pauli string given in the same form
                              as in the data structure of QubitOperator.
        updated_terms_1 (list): List of Pauli strings (QubitOperator),
                                which replace the original string if
                                they are shorter and they are equivalent
                                to each other.
        updated_terms_2 (list): List of Pauli strings given in the data
                                structure of QubitOperator, denoting which
                                strings the entries of the first list are
                                equivalent to.
    Returns:
        pauli_op (QubitOperator): Shortest Pauli string equivalent to the
                                  original.

    """
    pauli_op = QubitOperator(pauli_string)
    length = len(pauli_string)

    for x in numpy.arange(len(updated_terms_1)):
        if (pauli_string == updated_terms_2[x] and
            (length > len(list(updated_terms_1[x].terms)[0]))):
            pauli_op = updated_terms_1[x]
            length = len(list(updated_terms_1[x].terms)[0])
    return pauli_op


def _reduce_terms(terms, stabilizer_list, manual_input, fixed_positions):
    """
    Perform the term reduction using stabilizer conditions.

    Auxiliary function to reduce_number_of_terms.

    Args:
        terms (QubitOperator): Operator the number of terms is to be reduced.
        stabilizer_list (list): List of the stabilizers as QubitOperators.
        manual_input (Boolean): Option to pass the list of fixed qubits
                                positions manually. Set to False by default.
        fixed_positions (list): (optional) List of fixed qubit positions.
                                Passing a list is only effective if
                                manual_input is True.
    Returns:
        even_newer_terms (QubitOperator): Updated operator with reduced terms.
        fixed_positions (list): Positions of qubits to be used for the
                                term reduction.
    Raises:
        StabilizerError: Trivial stabilizer (identity).
        StabilizerError: Stabilizer with complex coefficient.
    """
    # Initialize fixed_position as an empty list to avoid conflict with
    # fixed_positions.
    if manual_input is False:
        fixed_positions = []

    # We need the index of the stabilizer to connect it to the fixed qubit.
    for i, _ in enumerate(stabilizer_list):
        selected_stab = list(stabilizer_list[0].terms)[0]

        if manual_input is False:
            # Find first position non-fixed position with non-trivial Pauli.
            for qubit_pauli in selected_stab:
                if qubit_pauli[0] not in fixed_positions:
                    fixed_positions += [qubit_pauli[0]]
                    fixed_op = qubit_pauli[1]
                    break

        else:
            # Finds Pauli of the fixed qubit.
            for qubit_pauli in selected_stab:
                if qubit_pauli[0] == fixed_positions[i]:
                    fixed_op = qubit_pauli[1]
                    break

        if fixed_op in ['X', 'Z']:
            other_op = 'Y'
        else:
            other_op = 'X'

        new_terms = QubitOperator()
        for qubit_pauli in terms:
            new_terms += fix_single_term(qubit_pauli, fixed_positions[i],
                                         fixed_op, other_op, stabilizer_list[0])
        updated_stabilizers = []
        for update_stab in stabilizer_list[1:]:
            updated_stabilizers += [
                fix_single_term(update_stab, fixed_positions[i], fixed_op,
                                other_op, stabilizer_list[0])
            ]

        # Update terms and stabilizer list.
        terms = new_terms
        stabilizer_list = updated_stabilizers

        check_stabilizer_linearity(stabilizer_list,
                                   msg='Linearly dependent stabilizers.')
        check_commuting_stabilizers(stabilizer_list,
                                    msg='Stabilizers anti-commute.')

    return terms, fixed_positions


def _reduce_terms_keep_length(terms, stabilizer_list, manual_input,
                              fixed_positions):
    """
    Perform the term reduction using stabilizer conditions.

    Auxiliary function to reduce_number_of_terms that returns the
    Pauli strings with the same length as in the starting operator.

    Args:
        terms (QubitOperator): Operator from which terms are reduced.
        stabilizer_list (list): List of the stabilizers as QubitOperators.
        manual_input (Boolean): Option to pass the list of fixed qubits
                                positions manually. Set to False by default.
        fixed_positions (list): (optional) List of fixed qubit positions.
                                Passing a list is only effective if
                                manual_input is True.
    Returns:
        even_newer_terms (QubitOperator): Updated operator with reduced terms.
        fixed_positions (list): Positions of qubits to be used for the
                                term reduction.
    Raises:
        StabilizerError: Trivial stabilizer (identity).
        StabilizerError: Stabilizer with complex coefficient.
    """
    term_list_duplicate = list(terms.terms)
    term_list = [QubitOperator(x) for x in term_list_duplicate]

    if manual_input is False:
        fixed_positions = []

    for i, x in enumerate(stabilizer_list):
        selected_stab = list(stabilizer_list[0].terms)[0]

        if manual_input is False:
            # Finds qubit position and its Pauli.
            for qubit_pauli in selected_stab:
                if qubit_pauli[0] not in fixed_positions:
                    fixed_positions += [qubit_pauli[0]]
                    fixed_op = qubit_pauli[1]
                    break
        else:
            # Finds Pauli of the fixed qubit.
            for qubit_pauli in selected_stab:
                if qubit_pauli[0] == fixed_positions[i]:
                    fixed_op = qubit_pauli[1]
                    break

        if fixed_op in ['X', 'Z']:
            other_op = 'Y'
        else:
            other_op = 'X'

        new_list = []
        updated_stabilizers = []
        for y in term_list:
            new_list += [
                fix_single_term(y, fixed_positions[i], fixed_op, other_op,
                                stabilizer_list[0])
            ]
        for update_stab in stabilizer_list[1:]:
            updated_stabilizers += [
                fix_single_term(update_stab, fixed_positions[i], fixed_op,
                                other_op, stabilizer_list[0])
            ]
        term_list = new_list
        stabilizer_list = updated_stabilizers

        check_stabilizer_linearity(stabilizer_list,
                                   msg='Linearly dependent stabilizers.')
        check_commuting_stabilizers(stabilizer_list,
                                    msg='Stabilizers anti-commute.')

    new_terms = QubitOperator()
    for x, ent in enumerate(term_list):
        new_terms += ent * terms.terms[term_list_duplicate[x]]
    for x, ent in enumerate(term_list):
        term_list_duplicate[x] = (QubitOperator(term_list_duplicate[x]) /
                                  list(ent.terms.items())[0][1])
        term_list[x] = list(ent.terms)[0]

    even_newer_terms = QubitOperator()
    for pauli_string, coefficient in new_terms.terms.items():
        even_newer_terms += coefficient * _lookup_term(
            pauli_string, term_list_duplicate, term_list)

    return even_newer_terms, fixed_positions


def reduce_number_of_terms(operator,
                           stabilizers,
                           maintain_length=False,
                           output_fixed_positions=False,
                           manual_input=False,
                           fixed_positions=None):
    r"""
    Reduce the number of Pauli strings of operator using stabilizers.

    This function reduces the number of terms in a string by merging
    terms that are identical by the multiplication of stabilizers.
    The resulting Pauli strings maintain their length, unless specified
    otherwise. In the latter case, a list of indices can be passed to
    manually indicate the qubits to be fixed.

    It is possible to reduce the number of terms in a Hamiltonian by
    merging Pauli strings :math:`H_1, \, H_2` that are related by a
    stabilizer :math:`S` such that  :math:`H_1 = H_2 \cdot S`. Given
    a stabilizer generator :math:`\pm X \otimes p` this algorithm fixes the
    first qubit, such that every Pauli string in the Hamiltonian acts with
    either :math:`Z` or the identity on it. Where necessary, this is achieved
    by multiplications with :math:`\pm X \otimes p`: a string
    :math:`Y \otimes h`, for instance,  is turned into
    :math:`Z \otimes (\mp ih\cdot p)`. Qubits on which a generator acts as
    :math:`Y` (:math:`Z`) are constrained to be acted on by the Hamiltonian as
    :math:`Z` (:math:`X`) or the identity. Fixing a different qubit for every
    stabilizer generator eliminates all redundant strings. The fixed
    representations are in the end re-expressed as the shortest of the
    original strings, :math:`H_1` or :math:`H_2`.


    Args:
        operator (QubitOperator): Operator of which the number of terms
                                  will be reduced.
        stabilizers (QubitOperator): Stabilizer generators used for the
                                     reduction. Can also be passed as a list
                                     of QubitOperator.
        maintain_length (Boolean): Option deciding whether the fixed Pauli
                                   strings are re-expressed in their original
                                   form. Set to False by default.
        output_fixed_positions (Boolean): Option deciding whether to return
                                          the list of fixed qubit positions.
                                          Set to False by default.
        manual_input (Boolean): Option to pass the list of fixed qubits
                                positions manually. Set to False by default.
        fixed_positions (list): (optional) List of fixed qubit positions.
                                Passing a list is only effective if
                                manual_input is True.
    Returns:
        reduced_operator (QubitOperator): Operator with reduced number of
                                          terms.

        fixed_positions (list): (optional) Fixed qubits.

    Raises:
        TypeError: Input terms must be QubitOperator.
        TypeError: Input stabilizers must be QubitOperator or list.
        StabilizerError: Trivial stabilizer (identity).
        StabilizerError: Stabilizer with complex coefficient.
        TypeError: List of fixed qubits required if manual input True.
        StabilizerError: The number of stabilizers must be equal to the number
                         of qubits manually fixed.
        StabilizerError: All qubit positions must be different.
    """
    if not isinstance(operator, QubitOperator):
        raise TypeError('Input terms must be QubitOperator.')
    if not isinstance(stabilizers, (QubitOperator, list, tuple, numpy.ndarray)):
        raise TypeError('Input stabilizers must be QubitOperator or list.')

    stabilizer_list = list(stabilizers)

    check_stabilizer_linearity(stabilizer_list,
                               msg='Trivial stabilizer (identity).')
    check_commuting_stabilizers(stabilizer_list,
                                msg='Stabilizer with complex coefficient.')

    if manual_input:
        # Convert fixed_position into a list to allow any type of
        # array_like data structure.
        if fixed_positions is None:
            raise TypeError('List of qubit positions required.')
        fixed_positions = list(fixed_positions)
        if len(fixed_positions) != len(stabilizer_list):
            raise StabilizerError('The number of stabilizers must be equal ' +
                                  'to the number of qubits manually fixed.')
        if len(set(fixed_positions)) != len(stabilizer_list):
            raise StabilizerError('All qubit positions must be different.')

    if maintain_length:
        (reduced_operator,
         fixed_positions) = _reduce_terms_keep_length(operator, stabilizer_list,
                                                      manual_input,
                                                      fixed_positions)
    else:
        (reduced_operator,
         fixed_positions) = _reduce_terms(operator, stabilizer_list,
                                          manual_input, fixed_positions)

    if output_fixed_positions:
        return reduced_operator, fixed_positions
    else:
        return reduced_operator


def taper_off_qubits(operator,
                     stabilizers,
                     manual_input=False,
                     fixed_positions=None,
                     output_tapered_positions=False):
    r"""
    Remove qubits from given operator.

    Qubits are removed by eliminating an equivalent number of
    stabilizer conditions. Which qubits that are can either be determined
    automatically or their positions can be set manually.

    Qubits can be disregarded from the Hamiltonian when the effect of all its
    terms on them is rendered trivial. This algorithm employs a stabilizers
    like :math:`\pm X \otimes p` to fix the action of every Pauli
    string on the first qubit to :math:`Z` or the identity. A string
    :math:`X \otimes h` would for instance be multiplied with the stabilizer
    to obtain :math:`1 \otimes (\pm h\cdot p)` while a string
    :math:`Z \otimes h^\prime` would pass without correction. The first
    qubit can subsequently be removed as it must be in the computational basis
    in Hamiltonian eigenstates.
    For stabilizers acting as :math:`Y` (:math:`Z`) on selected qubits,
    the algorithm would fix the action of every Hamiltonian string to
    :math:`Z` (:math:`X`). Updating also the list of remaining stabilizer
    generators, the algorithm is run iteratively.

    Args:
        operator (QubitOperator): Operator of which qubits will be removed.
        stabilizers (QubitOperator): Stabilizer generators for the tapering.
                                     Can also be passed as a list of
                                     QubitOperator.
        manual_input (Boolean): Option to pass the list of fixed qubits
                                positions manually. Set to False by default.
        fixed_positions (list): (optional) List of fixed qubit positions.
                                Passing a list is only effective if
                                manual_input is True.
        output_tapered_positions (Boolean): Option to output the positions of
                                            qubits that have been removed.
    Returns:
        skimmed_operator (QubitOperator): Operator with fewer qubits.
        removed_positions (list): (optional) List of removed qubit positions.
                                  For the qubits to be gone in the qubit count,
                                  the remaining qubits have been moved up to
                                  those indices.
    """
    if isinstance(stabilizers, (list, tuple, numpy.ndarray)):
        n_qbits_stabs = 0
        for ent in stabilizers:
            if op_utils.count_qubits(ent) > n_qbits_stabs:
                n_qbits_stabs = op_utils.count_qubits(ent)
    else:
        n_qbits_stabs = op_utils.count_qubits(stabilizers)

    n_qbits = max(op_utils.count_qubits(operator), n_qbits_stabs)

    (ham_to_update,
     qbts_to_rm) = reduce_number_of_terms(operator,
                                          stabilizers,
                                          maintain_length=False,
                                          manual_input=manual_input,
                                          fixed_positions=fixed_positions,
                                          output_fixed_positions=True)

    # Gets a list of the order of the qubits after tapering
    qbit_order = list(numpy.arange(n_qbits - len(qbts_to_rm), dtype=int))
    # Save the original list before it gets ordered
    removed_positions = qbts_to_rm
    qbts_to_rm.sort()
    for x in qbts_to_rm:
        qbit_order.insert(x, 'remove')

    # Remove the qubits
    skimmed_operator = QubitOperator()
    for term, coef in ham_to_update.terms.items():
        if term == ():
            skimmed_operator += QubitOperator('', coef)
            continue
        tap_tpls = []
        for p in term:
            if qbit_order[p[0]] != 'remove':
                tap_tpls.append((qbit_order[p[0]].item(), p[1]))

        skimmed_operator += QubitOperator(tuple(tap_tpls), coef)

    if output_tapered_positions:
        return skimmed_operator, removed_positions
    else:
        return skimmed_operator



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""A Trotter algorithm using a split-operator approach."""

from typing import cast, Optional, Sequence, Tuple

import cirq

# import openfermion.circuits.gates as gates
from openfermion.circuits.gates import rot11, rot111
import openfermion.ops as ops
# import openfermion.circuits.primitives as primitives
from openfermion.circuits.primitives import bogoliubov_transform, swap_network
from openfermion.circuits.trotter.trotter_algorithm import (Hamiltonian,
                                                            TrotterStep,
                                                            TrotterAlgorithm)


class SplitOperatorTrotterAlgorithm(TrotterAlgorithm):
    """A Trotter algorithm using a split-operator approach.

    This algorithm simulates a DiagonalCoulombHamiltonian. It uses Bogoliubov
    transformations to switch between a basis in which the one-body terms are
    convenient to simulate and a basis in which the two-body terms are
    convenient to simulate. The Bogoliubov transformations are implemented
    using Givens rotations.

    This algorithm is described in arXiv:1706.00023.
    """
    # TODO Maybe use FFFT

    supported_types = {ops.DiagonalCoulombHamiltonian}

    def symmetric(self, hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return SymmetricSplitOperatorTrotterStep(hamiltonian)

    def asymmetric(self, hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return AsymmetricSplitOperatorTrotterStep(hamiltonian)

    def controlled_symmetric(self,
                             hamiltonian: Hamiltonian) -> Optional[TrotterStep]:
        return ControlledSymmetricSplitOperatorTrotterStep(hamiltonian)

    def controlled_asymmetric(self, hamiltonian: Hamiltonian
                             ) -> Optional[TrotterStep]:
        return ControlledAsymmetricSplitOperatorTrotterStep(hamiltonian)


SPLIT_OPERATOR = SplitOperatorTrotterAlgorithm()


class SplitOperatorTrotterStep(TrotterStep):

    def __init__(self,
                 hamiltonian: 'openfermion.DiagonalCoulombHamiltonian') -> None:
        quad_ham = ops.QuadraticHamiltonian(hamiltonian.one_body)
        # Get the basis change matrix that diagonalizes the one-body term
        # and associated orbital energies
        self.orbital_energies, self.basis_change_matrix, _ = (
            quad_ham.diagonalizing_bogoliubov_transform())
        super().__init__(hamiltonian)


class SymmetricSplitOperatorTrotterStep(SplitOperatorTrotterStep):

    def prepare(self,
                qubits: Sequence[cirq.Qid],
                control_qubits: Optional[cirq.Qid] = None) -> cirq.OP_TREE:
        # Change to the basis in which the one-body term is diagonal
        yield cirq.inverse(
            bogoliubov_transform(qubits, self.basis_change_matrix))

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:

        n_qubits = len(qubits)

        # Simulate the one-body terms for half of the full time
        yield (cirq.rz(rads=-0.5 * self.orbital_energies[i] * time).on(
            qubits[i]) for i in range(n_qubits))

        # Rotate to the computational basis
        yield bogoliubov_transform(qubits, self.basis_change_matrix)

        # Simulate the two-body terms for the full time
        def two_body_interaction(p, q, a, b) -> cirq.OP_TREE:
            yield rot11(rads=-2 * self.hamiltonian.two_body[p, q] * time).on(
                a, b)

        yield swap_network(qubits, two_body_interaction)
        # The qubit ordering has been reversed
        qubits = qubits[::-1]

        # Rotate back to the basis in which the one-body term is diagonal
        yield cirq.inverse(
            bogoliubov_transform(qubits, self.basis_change_matrix))

        # Simulate the one-body terms for half of the full time
        yield (cirq.rz(rads=-0.5 * self.orbital_energies[i] * time).on(
            qubits[i]) for i in range(n_qubits))

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        # A Trotter step reverses the qubit ordering
        return qubits[::-1], None

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        # Rotate back to the computational basis
        yield bogoliubov_transform(qubits, self.basis_change_matrix)
        # If the number of Trotter steps is odd, possibly swap qubits back
        if n_steps & 1 and not omit_final_swaps:
            yield swap_network(qubits)


class ControlledSymmetricSplitOperatorTrotterStep(SplitOperatorTrotterStep):

    def prepare(self,
                qubits: Sequence[cirq.Qid],
                control_qubits: Optional[cirq.Qid] = None) -> cirq.OP_TREE:
        # Change to the basis in which the one-body term is diagonal
        yield cirq.inverse(
            bogoliubov_transform(qubits, self.basis_change_matrix))

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:

        n_qubits = len(qubits)

        if not isinstance(control_qubit, cirq.Qid):
            raise TypeError('Control qudit must be specified.')

        # Simulate the one-body terms for half of the full time
        yield (rot11(rads=-0.5 * self.orbital_energies[i] * time).on(
            control_qubit, qubits[i]) for i in range(n_qubits))

        # Rotate to the computational basis
        yield bogoliubov_transform(qubits, self.basis_change_matrix)

        # Simulate the two-body terms for the full time
        def two_body_interaction(p, q, a, b) -> cirq.OP_TREE:
            yield rot111(-2 * self.hamiltonian.two_body[p, q] * time).on(
                cast(cirq.Qid, control_qubit), a, b)

        yield swap_network(qubits, two_body_interaction)
        # The qubit ordering has been reversed
        qubits = qubits[::-1]

        # Rotate back to the basis in which the one-body term is diagonal
        yield cirq.inverse(
            bogoliubov_transform(qubits, self.basis_change_matrix))

        # Simulate the one-body terms for half of the full time
        yield (rot11(rads=-0.5 * self.orbital_energies[i] * time).on(
            control_qubit, qubits[i]) for i in range(n_qubits))

        # Apply phase from constant term
        yield cirq.rz(rads=-self.hamiltonian.constant * time).on(control_qubit)

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        # A Trotter step reverses the qubit ordering
        return qubits[::-1], control_qubit

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        # Rotate back to the computational basis
        yield bogoliubov_transform(qubits, self.basis_change_matrix)
        # If the number of Trotter steps is odd, possibly swap qubits back
        if n_steps & 1 and not omit_final_swaps:
            yield swap_network(qubits)


class AsymmetricSplitOperatorTrotterStep(SplitOperatorTrotterStep):

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:

        n_qubits = len(qubits)

        # Simulate the two-body terms for the full time
        def two_body_interaction(p, q, a, b) -> cirq.OP_TREE:
            yield rot11(rads=-2 * self.hamiltonian.two_body[p, q] * time).on(
                a, b)

        yield swap_network(qubits, two_body_interaction)
        # The qubit ordering has been reversed
        qubits = qubits[::-1]

        # Rotate to the basis in which the one-body term is diagonal
        yield cirq.inverse(
            bogoliubov_transform(qubits, self.basis_change_matrix))

        # Simulate the one-body terms for the full time
        yield (cirq.rz(rads=-self.orbital_energies[i] * time).on(qubits[i])
               for i in range(n_qubits))

        # Rotate back to the computational basis
        yield bogoliubov_transform(qubits, self.basis_change_matrix)

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        # A Trotter step reverses the qubit ordering
        return qubits[::-1], None

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        # If the number of Trotter steps is odd, possibly swap qubits back
        if n_steps & 1 and not omit_final_swaps:
            yield swap_network(qubits)


class ControlledAsymmetricSplitOperatorTrotterStep(SplitOperatorTrotterStep):

    def trotter_step(self,
                     qubits: Sequence[cirq.Qid],
                     time: float,
                     control_qubit: Optional[cirq.Qid] = None) -> cirq.OP_TREE:

        n_qubits = len(qubits)

        if not isinstance(control_qubit, cirq.Qid):
            raise TypeError('Control qudit must be specified.')

        # Simulate the two-body terms for the full time
        def two_body_interaction(p, q, a, b) -> cirq.OP_TREE:
            yield rot111(-2 * self.hamiltonian.two_body[p, q] * time).on(
                cast(cirq.Qid, control_qubit), a, b)

        yield swap_network(qubits, two_body_interaction)
        # The qubit ordering has been reversed
        qubits = qubits[::-1]

        # Rotate to the basis in which the one-body term is diagonal
        yield cirq.inverse(
            bogoliubov_transform(qubits, self.basis_change_matrix))

        # Simulate the one-body terms for the full time
        yield (rot11(rads=-self.orbital_energies[i] * time).on(
            control_qubit, qubits[i]) for i in range(n_qubits))

        # Rotate back to the computational basis
        yield bogoliubov_transform(qubits, self.basis_change_matrix)

        # Apply phase from constant term
        yield cirq.rz(rads=-self.hamiltonian.constant * time).on(control_qubit)

    def step_qubit_permutation(
            self,
            qubits: Sequence[cirq.Qid],
            control_qubit: Optional[cirq.Qid] = None
    ) -> Tuple[Sequence[cirq.Qid], Optional[cirq.Qid]]:
        # A Trotter step reverses the qubit ordering
        return qubits[::-1], control_qubit

    def finish(self,
               qubits: Sequence[cirq.Qid],
               n_steps: int,
               control_qubit: Optional[cirq.Qid] = None,
               omit_final_swaps: bool = False) -> cirq.OP_TREE:
        # If the number of Trotter steps is odd, possibly swap qubits back
        if n_steps & 1 and not omit_final_swaps:
            yield swap_network(qubits)
